#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" 
Collection of functions to convert audio and voltage data to Sound Pressure 
Level (SPL in Pascal) and Leq (Continuous Equivalent SPL).
Functions are adapted from the algorithm proposed in 
Merchant, N.D., Fristrup, K.M., Johnson, M.P., Tyack, P.L., Witt, M.J., Blondel, P. and Parks, S.E. (2015), 
Measuring acoustic habitats. Methods Ecol Evol, 6: 257-265. `DOI: 10.1111/2041-210X.12330 <https://doi.org/10.1111/2041-210X.12330>`_
"""   
#
# Authors:  Juan Sebastian ULLOA <lisofomia@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License 


#****************************************************************************
# -------------------       Load modules           ---------------------------
#****************************************************************************

# Import external modules

import numpy as np 
from numpy import sum, log10, abs, mean, sqrt

# min value
import sys
_MIN_ = sys.float_info.min

#****************************************************************************
# -------------------       Functions             ---------------------------
#****************************************************************************

#%%
def wav2volt (wave, Vadc=2): 
    """
    Convert an audio signal amplitude to Volts.

    .. warning::
       **Important:** Previous versions of this function incorrectly calculated the
        voltage conversion by omitting the division by 2 for the `Vadc` parameter.
        This resulted in voltage values that were twice as large as they should.
        If you want to correct the voltage values obtained with the previous version,
        you can simply divide the results by 2. 
        In dB SPL, this lead to a 6 dB increase in the calculated sound pressure level.
        If you want to correct the SPL values obtained with the previous version,
        you can simply subtract 6 dB from the old results.
        This has been corrected in the version 1.5.1. Ensure you are using the
        latest version (>=1.5.1) to obtain accurate voltage values.
    
    Parameters
    ----------
    wave : ndarray-like or scalar 
        wave should already be normalized between -1 to 1 (depending on the number of bits)
        take the output of the function sound.load of maad module
        ndarray-like or scalar containing the raw sound waveform 
        
    Vadc : scalar, optional, default is 2Vpp (=>+/-1V)
        Maximal voltage (peak to peak) converted by the analog to digital convertor ADC  
                
    Returns
    -------
    volt : ndarray-like or scalar
        ndarray-like or scalar containing the sound waveform in volt
        
    Examples
    --------
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> maad.spl.wav2volt(wave=w, Vadc=2)
        array([ 0.02155849,  0.02570888,  0.02583096, ..., -0.0082877 ,
        -0.00219072, -0.00377764])
    """
    # force to be ndarray
    wave = np.asarray(wave)
    
    volt =wave * Vadc/2
    return volt

#%%
def volt2pressure(volt, gain, sensitivity=-35, dBref=94):  
    """
    Convert Volts to instantaneous sound pressure (p [Pa]).
    
    Parameters
    ----------
    volt : ndarray-like or scalar
        ndarray-like or scalar containing the sound waveform in volt
        
    gain : integer
        Total gain applied to the sound (preamplifer + amplifier)
    
    sensitivity : float, optional, default is -35 (dB/V)
        Sensitivity of the microphone
    
    dBref : integer, optional, default is 94 (dBSPL)
        Pressure sound level used for the calibration of the microphone 
        (usually 94dB, sometimes 114dB)
                
    Returns
    -------
    p : ndarray-like or scalar
        ndarray-like or scalar containing the sound waveform in pressure (Pa)
        
    Examples
    --------
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> v = maad.spl.wav2volt(wave=w)
    >>> maad.spl.volt2pressure(volt=v, gain=42) 
        array([ 0.00481491,  0.00574187,  0.00576913, ..., -0.00185099,
        -0.00097856, -0.00168741])   
    
    Same result with the function wav2pressure
    
    >>> maad.spl.wav2pressure(wave=w, gain=42)
        array([ 0.00481491,  0.00574187,  0.00576913, ..., -0.00185099,
        -0.00097856, -0.00168741])
    
    """
    # force to be ndarray
    volt = np.asarray(volt)
    
    # wav to instantaneous sound pressure level (SPL)
    # coefficient to convert Volt into pressure (Pa)
    coeff = 1/10**(sensitivity/20) 
    p = volt * coeff / 10**(gain/20)
    return p

#%%
def wav2pressure (wave, gain, Vadc = 2, sensitivity=-35, dBref=94): 
    """
    Convert wave amplitude to instantaneous sound pressure (p [Pa]).

    .. warning::
       **Important:** Previous versions of this function incorrectly calculated the
        pressure conversion by omitting the division by 2 for the `Vadc` parameter.
        This resulted in pressure values that were twice as large as they should.
        If you want to correct the pressure values obtained with the previous version,
        you can simply divide the results by 2. 
        In dB SPL, this would lead to a 6 dB increase in the calculated sound pressure level.
        If you want to correct the SPL values obtained with the previous version,
        you can simply subtract 6 dB from the old results.
        This has been corrected in the version 1.5.1. Ensure you are using the
        latest version (>=1.5.1) to obtain accurate pressure values.
    
    Parameters
    ----------
    wave : ndarray-like or scalar 
        wave should already be normalized between -1 to 1 (depending on the number of bits)
        take the output of the function sound.load of maad module
        ndarray-like or scalar containing the raw sound waveform 
        
    gain : integer
        Total gain applied to the sound (preamplifer + amplifier)
        
    Vadc : scalar, optional, default is 2Vpp (=>+/-1V)
        Maximal voltage (peak to peak) converted by the analog to digital convertor ADC    

    sensitivity : float, optional, default is -35 (dB/V)
        Sensitivity of the microphone
    
    dBref : integer, optional, default is 94 (dBSPL)
        Pressure sound level used for the calibration of the microphone 
        (usually 94dB, sometimes 114dB)
                
    Returns
    -------
    p : ndarray-like or scalar
        ndarray-like or scalar containing the sound waveform in pressure (Pa)
        

    Examples
    --------
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> maad.spl.wav2pressure(wave=w, gain=42)
        array([ 0.00481491,  0.00574187,  0.00576913, ..., -0.00185099,
        -0.00097856, -0.00168741])
    
    Same result with 2 functions
    
    >>> v = maad.spl.wav2volt(wave=w)
    >>> maad.spl.volt2pressure(volt=v, gain=42) 
        array([ 0.00481491,  0.00574187,  0.00576913, ..., -0.00185099,
        -0.00097856, -0.00168741])        
    
    """
    # force to be ndarray
    wave = np.asarray(wave)

    v = wav2volt(wave, Vadc)
    p = volt2pressure(v, gain, sensitivity, dBref)
    return p

#%%
def pressure2dBSPL (p, pRef=20e-6): 
    """
    Convert sound pressure (p [Pa]) to sound pressure level (L [dB]).
    
    Parameters
    ----------
    p : ndarray-like or scalar
        Array or scalar containing the sound pressure in Pa 
                
    pRef : Sound pressure reference in the medium (air:20e-6 Pa, water:1e-6 Pa)
                
    Returns
    -------
    L : ndarray-like or scalar
        Array or scalar containing the sound pressure level (L [dB])
        

    Examples
    --------
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> p = maad.spl.wav2pressure(wave=w, gain=42)
    
    Get instantaneous sound pressure level (L)
    
    >>> maad.spl.pressure2dBSPL(abs(p))
        array([47.63116868, 49.16046498, 49.20160951, ..., 39.32747759,
        33.79115165, 38.52380598])
    
    Get equivalent sound pressure level (Leq) from the RMS of the pressure signal
    
    >>> p_rms = maad.util.rms(p)
    >>> maad.spl.pressure2dBSPL(p_rms)  
        48.51429086346293
        
    """    
    # force to be ndarray
    p = np.asarray(p)
    
    # if p ==0 set to MIN
    p[p==0] = _MIN_
    
    # Take the log of the ratio pressure/pRef
    L = 20*log10(p/pRef) 
    return L

#%%
def dBSPL2pressure (L, pRef=20e-6): 
    """
    Convert sound pressure level (L [dB]) to sound pressure (p [Pa]).
    
    Parameters
    ----------
    L : ndarray-like or scalar
        Array or scalar containing the sound pressure level (L [dB])
                
    pRef : Sound pressure reference in the medium (air:20e-6 Pa, water:1e-6 Pa)
                
    Returns
    -------
    p : ndarray-like or scalar
        Array or scalar containing the sound pressure in Pa 
        
    Examples
    --------
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> p = maad.spl.wav2pressure(wave=w, gain=42)
    >>> p_rms = maad.util.rms(p)
    >>> print(p_rms)
        0.0053302126891706676
    >>> L = maad.spl.pressure2dBSPL(p_rms)  
    >>> maad.spl.dBSPL2pressure(L)
        spl.dBSPL2pressure(L)
    """    
    # force to be ndarray
    L = np.asarray(L)
    
    # dB SPL to pressure
    p = 10**(L/20)*pRef
    return p

#%%
def wav2dBSPL (wave, gain, Vadc=2, sensitivity=-35, dBref=94, pRef=20e-6): 
    """
    Convert wave amplitude to instantaneous sound pressure level (L [dB SPL]).

    .. warning::
       **Important:** Previous versions of this function incorrectly calculated the
        pressure conversion by omitting the division by 2 for the `Vadc` parameter.
        In dB SPL, this would lead to a 6 dB increase in the calculated sound pressure level.
        If you want to correct the SPL values obtained with the previous version,
        you can simply subtract 6 dB from the old results.
        This has been corrected in the version 1.5.1. Ensure you are using the
        latest version (>=1.5.1) to obtain accurate dB SPL values.
    
    Parameters
    ----------
    wave : ndarray-like or scalar 
        wave should already be normalized between -1 to 1 (depending on the number of bits)
        take the output of the function sound.load of maad module
        ndarray-like or scalar containing the raw sound waveform 
        
    gain : integer
        Total gain applied to the sound (preamplifer + amplifier)
        
    Vadc : scalar, optional, default is 2Vpp (=>+/-1V)
        Maximal voltage (peak to peak) converted by the analog to digital convertor ADC     

    sensitivity : float, optional, default is -35 (dB/V)
        Sensitivity of the microphone
    
    dBref : integer, optional, default is 94 (dBSPL)
        Pressure sound level used for the calibration of the microphone 
        (usually 94dB, sometimes 114dB)
        
    pRef : Sound pressure reference in the medium (air:20e-6 Pa, water:1e-6 Pa)
                
    Returns
    -------
    L : ndarray-like or scalar
        ndarray-like or scalar containing the sound waveform in dB SPL (Sound Pressure level in dB)
        

    Examples
    --------
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> L = maad.spl.wav2dBSPL(wave=w, gain=42)
    
    Get an approximate of the equivalent sound pressure level (Leq)
        
    >>> maad.util.mean_dB(L)
        48.51429086346294
        
    Get equivalent sound pressure level (Leq) from the dedicated function
        
    >>> Leq = maad.spl.wav2leq(w, fs, gain=42, dt=1)
    >>> Leq_mean = maad.util.mean_dB(Leq)
    >>> Leq_mean
        48.55488267086038
    
    """       
    # force to be ndarray
    wave = np.asarray(wave)
    
    p = wav2pressure(wave, gain, Vadc, sensitivity, dBref)
    p_abs = abs(p)
    L = pressure2dBSPL(p_abs, pRef)
    return L

#%%
def amplitude2dBSPL (s, gain, Vadc=2, sensitivity=-35, dBref=94, pRef=20e-6): 
    """
    Convert signal (amplitude) to instantaneous sound pressure level (L [dB SPL]).

    .. warning::
       **Important:** Previous versions of this function incorrectly calculated the
        volt conversion by omitting the division by 2 for the `Vadc` parameter.
        In dB SPL, this would lead to a 6 dB increase in the calculated sound pressure level.
        If you want to correct the SPL values obtained with the previous version,
        you can simply subtract 6 dB from the old results.
        This has been corrected in the version 1.5.1. Ensure you are using the
        latest version (>=1.5.1) to obtain accurate dB SPL values.
    
    Parameters
    ----------
    s : ndarray-like or scalar 
        s is an amplitude signal (not energy signal : s² )
        
    gain : integer
        Total gain applied to the sound (preamplifer + amplifier)
        
    Vadc : scalar, optional, default is 2Vpp (=>+/-1V)
        Maximal voltage (peak to peak) converted by the analog to digital convertor ADC     

    sensitivity : float, optional, default is -35 (dB/V)
        Sensitivity of the microphone
    
    dBref : integer, optional, default is 94 (dBSPL)
        Pressure sound level used for the calibration of the microphone 
        (usually 94dB, sometimes 114dB)
        
    pRef : Sound pressure reference in the medium (air:20e-6 Pa, water:1e-6 Pa)
                
    Returns
    -------
    L : ndarray-like or scalar
        ndarray-like or scalar containing the sound waveform in dB SPL (Sound Pressure level in dB)
    
    See also
    --------
    power2dBSPL

    Examples
    --------
    >>> import numpy as np
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> Sxx_amplitude,tn,fn,_ = maad.sound.spectrogram (w, fs, nperseg=1024, mode='amplitude')  
    >>> S_amplitude_mean = np.mean(Sxx_amplitude, axis=1)
    
    Get instantaneous sound pressure level (L).
    
    >>> maad.spl.amplitude2dBSPL(S_amplitude_mean, gain=42)    
        array([ 33.54362056,  38.35057445,  35.93645686,  33.5848702 ,
                30.41177332,  27.22906075,  25.42079962,  24.42677998,
                22.89529647,  21.24902085,  20.75161999,  20.25268302,
                ...
                -12.18064124, -12.32827014, -12.48541472, -12.51967103,
                -12.54814367, -12.54536435, -12.75677825, -12.77745939])
    """       
    # force to be ndarray
    s = np.asarray(s)
    
    L = wav2dBSPL (s, gain, Vadc, sensitivity, dBref, pRef)

    return L


#%%
def power2dBSPL (P, gain, Vadc=2, sensitivity=-35, dBref=94, pRef=20e-6): 
    """
    Convert power (amplitude²) to sound pressure level (L [dB]).

    .. warning::
       **Important:** Previous versions of this function incorrectly calculated the
        volt conversion by omitting the division by 2 for the `Vadc` parameter.
        In dB SPL, this would lead to a 6 dB increase in the calculated sound pressure level.
        If you want to correct the SPL values obtained with the previous version,
        you can simply subtract 6 dB from the old results.
        This has been corrected in the version 1.5.1. Ensure you are using the
        latest version (>=1.5.1) to obtain accurate dB SPL values.    
    
    Parameters
    ----------
    P : ndarray-like or scalar
        ndarray-like or scalar containing the power signal (P), for instance 
        Sxx_power, the power spectral density (PSD)

    gain : integer
        Total gain applied to the sound (preamplifer + amplifier)
        
    Vadc : scalar, optional, default is 2Vpp (=>+/-1V)
        Maximal voltage (peak to peak) converted by the analog to digital convertor ADC     

    sensitivity : float, optional, default is -35 (dB/V)
        Sensitivity of the microphone
    
    dBref : integer, optional, default is 94 (dBSPL)
        Pressure sound level used for the calibration of the microphone 
        (usually 94dB, sometimes 114dB)
                
    pRef : Sound pressure reference in the medium (air:20e-6 Pa, water:1e-6 Pa)
                
    Returns
    -------
    L : ndarray-like or scalar
        ndarray-like or scalar containing the sound pressure level (L [dB])
        
    See also
    --------
    amplitude2dBSPL
        
    Examples
    --------
    >>> import numpy as np
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> Sxx_power,tn,fn,_ = maad.sound.spectrogram (w, fs, nperseg=1024, mode='psd')  
    >>> S_power_mean = np.mean(Sxx_power, axis=1)
    
    Get instantaneous sound pressure level (L).
    
    >>> maad.spl.power2dBSPL(S_power_mean, gain=42)    
        array([ 3.55439604e+01,  3.94219755e+01,  3.71509454e+01,  3.54860553e+01,
                3.20633292e+01,  2.85071055e+01,  2.65508217e+01,  2.56607733e+01,
                2.43080132e+01,  2.24405108e+01,  2.18647044e+01,  2.14653511e+01,
                ...
                -1.11242771e+01, -1.12703297e+01, -1.14085678e+01, -1.14487227e+01,
                -1.14906343e+01, -1.14980091e+01, -1.16907591e+01, -1.17027481e+01])
    """    
    # force to be ndarray
    P = np.asarray(P)
    # convert power (energy) to amplitude
    w = sqrt(P)
    # convert amplitude to dB sPL
    L = wav2dBSPL (w, gain, Vadc, sensitivity, dBref, pRef)
    
    return L

################################## Leq ########################################
#%%
def wav2leq (wave, f, gain, Vadc=2, dt=1, sensitivity=-35, dBref = 94, pRef = 20e-6): 
    """
    Convert wave to Equivalent Continuous Sound Pressure level (Leq [dB SPL]).

    .. warning::
       **Important:** Previous versions of this function incorrectly calculated the
        volt conversion by omitting the division by 2 for the `Vadc` parameter.
        In dB SPL, this would lead to a 6 dB increase in the calculated sound pressure level.
        If you want to correct the SPL values obtained with the previous version,
        you can simply subtract 6 dB from the old results.
        This has been corrected in the version 1.5.1. Ensure you are using the
        latest version (>=1.5.1) to obtain accurate dB SPL values. 
    
    Parameters
    ----------
    wave : ndarray-like or scalar 
        wave should already be normalized between -1 to 1 (depending on the number of bits)
        take the output of the function sound.load of maad module
        ndarray-like or scalar containing the raw sound waveform 
        
    f : integer
        Sampling frequency in Hz
        
    gain : integer
        Total gain applied to the sound (preamplifer + amplifier)
        
    Vadc : scalar, optional, default is 2Vpp (=>+/-1V)
        Maximal voltage (peak to peak) converted by the analog to digital convertor ADC  
        
    dt : float, optional, default is 1 (second)
        Integration step to compute the Leq (Equivalent Continuous Sound level)

    sensitivity : float, optional, default is -35 (dB/V)
        Sensitivity of the microphone
    
    dBref : integer, optional, default is 94 (dBSPL)
        Pressure sound level used for the calibration of the microphone 
        (usually 94dB, sometimes 114dB)
    
    pRef : Sound pressure reference in the medium (air:20e-6, water:1e-6 Pa)

    Returns
    -------
    Leq : float
        Equivalent Continuous Sound pressure level (Leq [dB SPL])

    Examples
    --------
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> Leq = maad.spl.wav2leq (w, fs, gain=42)  
    >>> maad.util.mean_dB(Leq)
        48.55488267086038
    """    
    # force to be ndarray
    wave = np.asarray(wave)
    
    # convert in Volt
    volt = wav2volt(wave, Vadc)
    ##### volt amplitude to Leq 
    # wav to RMS
    dN = int(np.floor(dt*f)) # RMS period in number of points
    N_RMS = int(np.floor(len(volt)/dN)) # number of RMS periods
    # reduction of volt vector length
    volt_1D = volt[0:dN*N_RMS]
    # reshape into 2D (the duration of each row is dt)
    volt_2D = volt_1D.reshape(N_RMS,dN)
    # RMS
    volt_RMS = sqrt(mean(volt_2D**2,axis=1))
    # if volt_RMS ==0 set to MIN
    volt_RMS[volt_RMS==0] = _MIN_
    # RMS to Leq (Equivalent Continuous Sound level)
    Leq = 20*log10(volt_RMS) - sensitivity - gain + dBref - 20*np.log10(pRef/20e-6)
    return Leq

#%%
def pressure2leq (p, fs, dt=1, pRef = 20e-6): 
    """
    Convert pressure vector (p [Pa]) to Equivalent Continuous Sound Pressure 
    level (Leq [dB SPL]).
    
    Parameters
    ----------
    p : ndarray-like
        Array containing the sound pressure in Pa 
        
    fs : integer
        Sampling frequency in Hz
        
    dt : float, optional, default is 1 (second)
        Integration step to compute the Leq (Equivalent Continuous Sound level)

    pRef : Sound pressure reference in the medium (air:20e-6, water:1e-6 Pa)
    
    Returns
    -------
    Leq : float
        Equivalent Continuous Sound pressure level (Leq [dB SPL])
        
    Examples
    --------
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> p = maad.spl.wav2pressure(wave=w, gain=42)
    >>> Leq = maad.spl.pressure2leq (p, fs)
    >>> maad.util.mean_dB(Leq)
        48.534282757580755
        
    """     
    # be sure they are ndarray
    p = np.asarray(p)
    
    ##### wav SPLto Leq 
    # wav to RMS
    dN = int(np.floor(dt*fs)) # RMS period in number of points
    N_RMS = int(np.floor(len(p)/dN)) # number of RMS periods
    # reduction of volt vector length
    p_1D = p[0:dN*N_RMS]
    # reshape into 2D (the duration of each row is dt)
    p_2D = p_1D.reshape(N_RMS,dN)
    # RMS
    p_RMS = sqrt(mean(p_2D**2,axis=1))
    # if p_RMS ==0 set to MIN
    p_RMS[p_RMS==0] = _MIN_
    # RMS to Leq (Equivalent Continuous Sound level)
    Leq = 20*log10(p_RMS/pRef)
    return Leq

#%%
def psd2leq (P, gain, Vadc=2, sensitivity=-35, dBref = 94, pRef = 20e-6):
    """
    Convert Power spectral density (PSD, amplitude²) into 
    Equivalent Continuous Sound pressure level (Leq [dB SPL])

    .. warning::
       **Important:** Previous versions of this function incorrectly calculated the
        volt conversion by omitting the division by 2 for the `Vadc` parameter.
        In dB SPL, this would lead to a 6 dB increase in the calculated sound pressure level.
        If you want to correct the SPL values obtained with the previous version,
        you can simply subtract 6 dB from the old results.
        This has been corrected in the version 1.5.1. Ensure you are using the
        latest version (>=1.5.1) to obtain accurate dB SPL values. 
    
    Parameters
    ----------
    P : ndarray-like (1d)
        ndarray-like containing the Power spectral density (PSD=amplitude²) 
        
    gain : integer
        Total gain applied to the sound (preamplifer + amplifier)
        
    Vadc : scalar, optional, default is 2Vpp (=>+/-1V)
        Maximal voltage (peak to peak) converted by the analog to digital convertor ADC  

    sensitivity : float, optional, default is -35 (dB/V)
        Sensitivity of the microphone
    
    dBref : integer, optional, default is 94 (dBSPL)
        Pressure sound level used for the calibration of the microphone 
        (usually 94dB, sometimes 114dB)

    pRef : Sound pressure reference in the medium (air:20e-6, water:1e-6 Pa)

    Returns
    -------
    Leq : float
        Equivalent Continuous Sound pressure level (Leq [dB SPL])
        
    Examples
    --------
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> Sxx_power,tn,fn,_ = maad.sound.spectrogram (w, fs)
    >>> S_power_mean = maad.sound.avg_power_spectro(Sxx_power) 
    >>> maad.spl.psd2leq(S_power_mean, gain=42)
        47.537824826354665
        
    """  
    # be sure they are ndarray
    P = np.asarray(P)
    
    # convert P (amplitude²) to pressure²
    P = wav2pressure (sqrt(P), gain, Vadc, sensitivity, dBref)**2
    
    # if P ==0 set to MIN
    P[P==0] = _MIN_
    
    # Energy (pressure^2) to Leq (=> Leq (Equivalent Continuous Sound level) 
    # if the sum is performed on the whole PSD)
    Leq = 10*log10(sum(P)/pRef**2)
    return Leq
    
################################ TEST ########################################
    
## Test the current operating system
if __name__ == "__main__":
    # ############## Import MAAD module
    from pathlib import Path # in order to be Windows/Linux/MacOS compatible
    import os
    
    # change the path to the current path where the script is located
    # Get the current dir of the current file
    dir_path = os.path.dirname(os.path.realpath('__file__'))
    os.chdir(dir_path)
    
    # data directory 
    datadir = (Path(dir_path).parents[1] / "data/").resolve()
    filename = datadir/"cold_forest_daylight.wav"
    
    # Go to the root directory where is the module maad and import maad
    maad_path = Path(dir_path).parents[1]
    os.sys.path.append(maad_path.as_posix())
    from maad import sound, util, spl
    
    ####### Variables
    SENSITIVITY     = -35       # Sensbility microphone (for SM4 it is -35dBV)
    GAIN_TOTAL      = 42        # total amplification gain : preamplifier gain = 26dB and gain = 16dB
    PREF            = 20e-6     # Reference pressure (in the air : 20µPa)
    DELTA_T         = 1         # Leq integration time dt in s
    NFFT            = 1024      # length of the fft
    VADC            = 2         # Maximal voltage (peak to peak) converted by the analog to digital convertor ADC (i.e : +/- 1V peak to peak => Vadc=2V)
    
    # Load the sound
    wav,fs = sound.load(filename=filename, channel='right', detrend=False, verbose=False)
    
    ## conversion into dB SPL 

    print('-------------------------------------------------------------------------')
    print('Leq calculation directly from the sound file (in time or frequency domain)')

    # convert sounds (wave) into SPL and subtract DC offset
    wav = wav - np.mean(wav)
    # wav -> pressure -> Leq
    p = spl.wav2pressure(wav, gain=GAIN_TOTAL, Vadc=VADC, sensitivity=SENSITIVITY, dBref=94)
    Leq = spl.pressure2leq(p, fs, dt=DELTA_T, pRef=PREF)
    print('Leq from volt', util.mean_dB(Leq))

    # wav -> Leq
    wav_Leq2 = spl.wav2leq(wav, f=fs, gain=GAIN_TOTAL, Vadc=VADC, dt=DELTA_T, 
                    sensitivity=SENSITIVITY, dBref=94, pRef=PREF) 
    print('Leq from wav', util.mean_dB(wav_Leq2))

    # Power Density Spectrum : PSD
    # with wav
    from numpy import fft
    P = abs(fft.fft(wav)/len(wav))**2
    # average Leq from PSD
    P_Leq3  = spl.psd2leq(P, gain=GAIN_TOTAL, Vadc=VADC, 
                    sensitivity=SENSITIVITY, dBref=94, pRef=PREF)
    print('Leq from PSD',P_Leq3)

    ##
    # Leq from spectrogram Sxx_power : 
    print('-------------------------------------------------------------------------')
    print('Leq calculation from Spectrogram (time-frequency representation of a sound)')

    # Power Density Spectrogram : Sxx_power
    # with wav
    Sxx_power,tn,fn,_ = sound.spectrogram (wav, fs=fs, nperseg=NFFT, mode='psd')  
    # Sxx_power to S_power
    S_power = np.mean(Sxx_power,axis=1)
    # average Leq from S_power
    P_Leq5  = spl.psd2leq(S_power, gain=GAIN_TOTAL, Vadc=VADC, 
                    sensitivity=SENSITIVITY, dBref=94, pRef=PREF) 
    print('Leq from Sxx_power spectrogram',P_Leq5)

    # total energy from S_power
    energy  = np.sum(S_power)
    P_Leq6  = spl.wav2dBSPL(np.sqrt(energy), gain=GAIN_TOTAL, Vadc=VADC, 
                    sensitivity=SENSITIVITY, dBref=94, pRef=PREF)
    print('Leq from S_power energy',P_Leq6)

    print('')
    print('> The difference with previous Leq calculation is due to the average of '+ 
            'the Sxx_power along the time axis which reduces the noise contribution into the total energy.')
    print('> By increasing the NFFT length (i.e. NFFT=4096), Leq value converges towards previous Leq values') 



    
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Ensemble of usefull functions for scikit-maad. """
#
# Authors:  Juan Sebastian ULLOA 
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

# =============================================================================
# Load the modules
# =============================================================================
# Import external modules
import matplotlib.pyplot as plt
import numpy as np
from numpy import log10, diff, mean
import pandas as pd
import numbers
import warnings

# min value
import sys
_MIN_ = sys.float_info.min

# %%


def index_bw(fn, bw):
    """
    Select all the index coresponding to a selected frequency band.

    Parameters
    ----------
    fn :  1d ndarray of scalars
        Vector of frequencies

    bw : single value or tupple of two values
        if single value : frequency to select
        if tupple of two values : min frequency and max frequency to select

    Returns
    -------
    index : 1d ndarray of scalars
        Vector of booleans corresponding to the selected frequency(-ies)

    Examples
    --------
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> Sxx_power,tn,fn,_ = maad.sound.spectrogram(w,fs,window='hann',noverlap=512, nFFT=1024)
    >>> Sxx_dB = maad.util.power2dB(Sxx_power) # convert into dB
    >>> bw = (2000,6000) #in Hz
    >>> fig_kwargs = {'vmax': Sxx_dB.max(),
                      'vmin': -90,
                      'extent':(tn[0], tn[-1], fn[0], fn[-1]),
                      'figsize':(10,13),
                      'title':'Power Spectrum Density (PSD)',
                      'xlabel':'Time [sec]',
                      'ylabel':'Frequency [Hz]',
                      }
    >>> maad.util.plot2d(Sxx_dB,**fig_kwargs)
    >>> Sxx_dB_crop = Sxx_dB[maad.util.index_bw(fn, bw)]
    >>> fn_crop = fn[maad.util.index_bw(fn, bw)]
    >>> fig_kwargs = {'vmax': Sxx_dB.max(),
                      'vmin': -90,
                      'extent':(tn[0], tn[-1], fn_crop[0], fn_crop[-1]),
                      'figsize':(10*len(fn_crop)/len(fn),13),
                      'title':'Power Spectrum Density (PSD)',
                      'xlabel':'Time [sec]',
                      'ylabel':'Frequency [Hz]',
                      }
    >>> maad.util.plot2d(Sxx_dB_crop,**fig_kwargs)

    """
    # select the indices corresponding to the frequency bins range
    if bw is None:
        index = (np.ones(len(fn)))
        index = [bool(x) for x in index]
    elif isinstance(bw, tuple) or isinstance(bw, np.ndarray):
        index = (fn >= fn[(abs(fn-bw[0])).argmin()]) * \
            (fn <= fn[(abs(fn-bw[1])).argmin()])
    elif isinstance(bw, numbers.Number):
        index = np.zeros(len(fn))
        index[(abs(fn-bw)).argmin()] = 1
        index = [bool(x) for x in index]
    return index

# %%

def into_bins(x, an, bin_step, axis=0, bin_min=None, bin_max=None, display=False):
    """ 
    Divide a vector (1D) or a matrix (2D) into multiple bins according to a 
    bin_step with respect of the energy

    Parameters
    ----------
    x : array-like
        1D or 2D array to bin
    an :1d ndarray of floats 
        Vector containing the positions of each value. 
        In case of 2D matrix, this vector corresponds to the horizontal (row)
        or vertical (columns) units
    bin_step : scalar
        Determine the width of each bin.
    axis : integer, optional, default is 0
        Determine  along which axis the transformation is done.
        In case of matrix :
        axis = 0 => transformation is done on column
        axis = 1 => transformation is done on row 
    bin_min : scalar, optional, default is None
        This minimum value corresponds to the start of the first bin. 
        By default, the minimum value is the first value of an.
    bin_max : scalar, optional, default is None
        This maximum value corresponds to end of the last bin. 
        By default, the maximum value is the last value of an.   
    display : boolean, optional, defualt is False
        Display the result of the tranformation : an histogram
        In case of matrix, the mean histogram is shown.

    Returns
    -------
    xbins :  array-like
        1D or 2D array which correspond to the data after being transformed into bins
    bin : 1d ndarray of floats 
        Vector containing the positions of each bin

    Examples
    --------
    >>> import numpy as np
    >>> frequency = np.arange(0,20000,100)
    >>> delta = np.random.uniform(-10,10, size=(len(frequency),))
    >>> amplitude = 0.001 * frequency + 3 + delta
    >>> fig_kwargs = {'figtitle':'',
                  'xlabel':'Frequency [Hz]',
                  'ylabel':'AU',
                  'linewidth': 0.5,
                  'legend' : 'before binning'
                  }
    >>> ax,fig = maad.util.plot1d(frequency,amplitude,**fig_kwargs)
    >>> amplitude_bin,frequency_bin = maad.util.into_bins(amplitude,frequency,bin_step=1000) 
    >>> fig_kwargs = {'figtitle':'',
                  'xlabel':'Frequency [Hz]',
                  'ylabel':'AU',
                  'linewidth': 0.5,
                  'color' :'red',
                  'legend' : 'after binning'
                  }
    >>> ax,fig = maad.util.plot1d(frequency_bin, amplitude_bin, ax, **fig_kwargs)

    """
    # force x to be an ndarray
    x = np.asarray(x)

    # Test if the bin_step is larger than the resolution of an
    if bin_step < (an[1]-an[0]):
        raise Exception(
            'WARNING: bin step must be larger or equal than the actual resolution of x')

    # In case the limits of the bin are not set
    if bin_min == None:
        bin_min = an[0]
    if bin_max == None:
        bin_max = an[-1]

    # Creation of the bins
    bins = np.arange(bin_min, bin_max+bin_step, bin_step)

    # select the indices corresponding to the frequency bins range
    b0 = bins[0]
    xbin = []
    s = []
    for index, b in enumerate(bins[1:]):
        indices = (an >= b0)*(an < b)
        s.append(sum(indices))
        if (axis is None) or (x.ndim == 1):
            xbin.append(mean(x[indices]))
        elif axis == 0:
            xbin.append(mean(x[indices, :], axis=axis))
        elif axis == 1:
            xbin.append(mean(x[:, indices], axis=axis))
        b0 = b

    xbin = np.asarray(xbin) * mean(s)
    bins = bins[0:-1]

    # Display
    if display:
        plt.figure()
        # if xbin is a vector
        if xbin.ndim == 1:
            plt.plot(an, x)
            plt.bar(bins, xbin, bin_step*0.75, alpha=0.5, align='edge')
        else:
            # if xbin is a matrix
            if axis == 0:
                axis = 1
            elif axis == 1:
                axis = 0
            plt.plot(an, mean(x, axis=axis))
            plt.bar(bins, mean(xbin, axis=1), bin_step *
                    0.75, alpha=0.5, align='edge')

    return xbin, bins

# %%


def rle(x):
    """
    Compute the Run-Length encoding (RLE) of a vector.

    Run-length encoding is a lossless data compression in which runs of data 
    (sequences in which the same data value occurs in many consecutive data elements) 
    are stored as a single data value and count. This is useful on data 
    that contains many repeated values, for example, simple graphic 
    images such as icons, line drawings, and animations. It is not useful with files that 
    don't have many runs as it could greatly increase the file size [1]_. 

    Parameters
    ----------
    x : 1D array
        vector with numbers (ndarray or list)

    Returns
    -------
    lengths, values : 1D array
        2 vectors that stores values (in values) and the number of times each
        value is repeated (in lengths).

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Run-length_encoding

    Examples
    --------
    RLE compression of the vector [1, 1, 1, 1, 2, 2, 3, 3, 3, 3, 4, 5, 5]

    >>> from maad.util import rle
    >>> length, values = rle([1, 1, 1, 2, 2, 3, 3, 3, 3, 4, 5, 5])
    >>> print('length:', length, '- values:', values)
    length: [3 2 4 1 2] - values: [1 2 3 4 5]

    """
    x = np.asarray(x)
    if x.ndim > 1:
        print("x must be a vector")
    else:
        n = len(x)
        states = x[1:] != x[:-1]
        i = np.r_[np.where(states)[0], n-1]
        lengths = np.r_[i[0]+1, diff(i)]
        values = x[i]
    return lengths, values

# %%


def linear_scale(x, minval=0.0, maxval=1.0, axis=0):
    """ 
    Scale the values of a vector or matrix from a user specified 
    minimum to a user specified maximum.

    Parameters
    ----------
    x : array-like
        numpy.array like with numbers, list or dataframe
    minval : scalar, optional, default : 0
        This minimum value is attributed to the minimum value of 
        - the array if axis = None
        - each column if axis =0
        - each row if axis = 1
    maxval : scalar, optional, default : 1
        This maximum value is attributed to the maximum value of  
        - the array if axis = None
        - each column if axis =0
        - each row if axis = 1
    axis : integer, optional, default : 0
        select if the min,max is calculated on the entire matrix (axis=None),
        on each column (axis=0), on each row (axis=1)

    Returns
    -------
    y : array-like
        numpy.array like with numbers  or dataframe

    References
    ----------
    Written by Aniruddha Kembhavi, July 11, 2007 for MATLAB, adapted by S. Haupert 
    Dec 12, 2017 for Python


    Examples
    --------
    >>> a = np.array([1,2,3,4,5])
    >>> maad.util.linear_scale(a, 0, 1)
        array([0.  , 0.25, 0.5 , 0.75, 1.  ])

    """

    # if x is a list, convert x into ndarray
    if isinstance(x, list):
        x = np.asarray(x)

    if isinstance(x, np.ndarray):
        # avoid problems with inf and -inf values
        x_min = np.min(x.ravel()[x.ravel() != -np.inf])
        x[np.where(x == -np.inf)] = x_min
        x_max = np.max(x.ravel()[x.ravel() != np.inf])
        x[np.where(x == np.inf)] = x_max

    # do the normalization
    y = x - x.min(axis)
    y = (y/y.max(axis))*(maxval-minval)
    y = y + minval
    return y

# %%


def amplitude2dB(x, db_range=None, db_gain=0):
    """
    Transform amplitude data (signal, scalar) into decibel scale within the 
    dB range (db_range). A gain (db_gain) could be added at the end.    

    Parameters
    ----------
    x : array-like or scalar
        data to rescale in dB  
    db_range : scalar, optional, default : None
        if db_range is a number, anything lower than -db_range is set to 
        -db_range and anything larger than 0 is set to 0
    db_gain : scalar, optional, default is 0
        Gain added to the results 
        amplitude --> 20*log10(x) + db_gain  

    Returns
    -------
    y : array-like or scalar
        y = 20*log10(x) + db_gain  

    Examples
    --------
    >>> a = np.array([1,2,3,4,5])
    >>> maad.util.amplitude2dB(a)
        array([ 0.        ,  6.02059991,  9.54242509, 12.04119983, 13.97940009])

    """
    x = abs(x)   # take the absolute value of x

    # Avoid zero value for log10
    # if it's a scalar
    if hasattr(x, "__len__") == False:
        if x == 0:
            x = _MIN_
    else:
        x[x == 0] = _MIN_  # Avoid zero value for log10

    # conversion in dB
    y = 20*log10(x)   # take log

    if db_gain:
        y = y + db_gain    # Add gain if needed

    if db_range is not None:
        # set anything above db_range as 0
        y[y > 0] = 0
        # set anything less than -db_range as -db_range
        y[y < -(db_range)] = -db_range

    return y

# %%


def power2dB(x, db_range=None, db_gain=0):
    """
    Transform power (amplitude²) signal or scalar into decibel scale 
    within the dB range (db_range). A gain (db_gain) can be added at the end.    

    Parameters
    ----------
    x : array-like
        data to rescale in dB  
    db_range : scalar, optional, default : None
        if db_range is a number, anything lower than -db_range is set to 
        -db_range and anything larger than 0 is set to 0
    db_gain : scalar, optional, default is 0
        Gain added to the results 
        amplitude --> 20*log10(x) + db_gain  

    Returns
    -------
    y : array-like or scalar
        y = 10*log10(x) + db_gain  

    Examples
    --------
    >>> a = np.array([1,2,3,4,5])
    >>> maad.util.power2dB(a**2)
        array([ 0.        ,  6.02059991,  9.54242509, 12.04119983, 13.97940009])

    """
    x = abs(x)   # take the absolute value of x

    # Avoid zero value for log10
    # if it's a scalar
    if hasattr(x, "__len__") == False:
        if x == 0:
            x = _MIN_
    else:
        x[x == 0] = _MIN_  # Avoid zero value for log10

    # conversion in dB
    y = 10*log10(x)   # take log

    if db_gain:
        y = y + db_gain    # Add gain if needed

    if db_range is not None:
        # set anything above db_range as 0
        y[y > 0] = 0
        # set anything less than -db_range as -db_range
        y[y < -(db_range)] = -db_range

    return y

# %%


def dB2amplitude(x, db_gain=0):
    """
    Transform data in dB scale into amplitude
    A gain (db_gain) could be added at the end.    

    Parameters
    ----------
    x : array-like or scalar
        data in dB to rescale in amplitude 
    db_gain : scalar, optional, default is 0
        Gain that was added to the result 
        --> 20*log10(x) + db_gain

    Returns
    -------
    y : array-like or scalar
        output in amplitude unit

    Examples
    --------
    >>> a = np.array([ 0.        ,  6.02059991,  9.54242509, 12.04119983, 13.97940009])
    >>> maad.util.dB2amplitude(a)
        array([1., 2., 3., 4., 5.])

    """
    y = 10**((x - db_gain)/20)

    return y

# %%


def dB2power(x, db_gain=0):
    """
    Transform data in dB scale into power (amplitude²)
    A gain (db_gain) could be added at the end.    

    Parameters
    ----------
    x : array-like or scalar
        data in dB to rescale in power 
    db_gain : scalar, optional, default is 0
        Gain that was added to the result 
        --> 10*log10(x) + db_gain

    Returns
    -------
    y : array-like or scalar
        output in power unit

    Examples
    --------
    >>> a = np.array([ 0.        ,  6.02059991,  9.54242509, 12.04119983, 13.97940009])
    >>> maad.util.dB2power(a)
        array([ 1.        ,  4.        ,  8.99999999, 16.00000001, 25.00000002])        
    """
    y = 10**((x - db_gain)/10)

    return y

# %%


def add_dB(*argv, axis=0):
    """
    Computes an addition on decibel values.

    Parameters
    ----------
    *argv : ndarray-like of floats
        Arrays containing the sound waveform in dB 

    axis : integer, optional, default is 0
        if addition of multiple arrays, select the axis on which the sum is done

    Returns
    -------
    e_sum : ndarray-like of floats
        Array containing the sum of the dB values

    Examples
    --------

    Example with an audio file

    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> Sxx_power,tn,fn,_ = maad.sound.spectrogram(w,fs)
    >>> L = maad.spl.power2dBSPL(Sxx_power,gain=42)
    >>> L_sum = maad.util.add_dB(L, axis=1)  
    >>> fig_kwargs = {'figtitle':'Spectrum (PSD)',
                      'xlabel':'Frequency [Hz]',
                      'ylabel':'Power [dB]',
                      }
    >>> fig, ax = maad.util.plot1d(fn, L_sum.transpose(), **fig_kwargs)

    Example with single values

    >>> L1 = 90 # 90dB
    >>> maad.util.add_dB(L1,L1)
        93.01029995663981

    Example with arrays

    >>> L1 = [90,80,70]
    >>> maad.util.add_dB(L1,L1,axis=1)
        array([90.45322979, 90.45322979])
    >>> maad.util.add_dB(L1,L1,axis=0)
        array([93.01029996, 83.01029996, 73.01029996])
    """
    # force to be ndarray
    L = np.asarray(argv)

    # test if the 1st dim is = 1 (in case argv is a single 2d array) in order to remove this dimension
    if L.shape[0] == 1:
        L = L[0, ]

    # Verify the adequation between axis number and number of dimensions of L
    if axis >= L.ndim:
        axis = L.ndim - 1

    # dB to energy as sum has to be done with energy
    e = dB2power(L)
    e_sum = e.sum(axis)

    # test if 2 dimensions but length on 1 dimension is 1
    if e_sum.ndim == 2:
        if (e_sum.shape[0] == 1) or (e_sum.shape[1] == 1):
            e_sum = np.ndarray.flatten(e_sum)

    # energy=>pressure to dB
    L_sum = power2dB(e_sum)

    return L_sum

# %%


def mean_dB(*argv, axis=0):
    """
    Compute the average of decibel values.

    Parameters
    ----------
    *argv : ndarray-like of floats
        Arrays containing the sound waveform in dB

    axis : integer, optional, default is 0
        if average of multiple arrays, select the axis on which the average is done

    Returns
    -------
    e_mean : ndarray-like of floats
        Array containing the mean of the dB values

    Examples
    --------

    Example with an audio file

    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> Sxx_power,tn,fn,_ = maad.sound.spectrogram(w,fs)
    >>> L = maad.util.power2dBSPL(Sxx_power,gain=42)
    >>> L_mean = maad.util.mean_dB(L, axis=1)  
    >>> fig_kwargs = {'figtitle':'Power spectrum (PSD)',
                      'xlabel':'Frequency [Hz]',
                      'ylabel':'Power [dB]',
                      }
    >>> fig, ax = maad.util.plot1d(fn, L_mean.transpose(), **fig_kwargs)

    Example with single values

    >>> L1 = 90 # 90dB
    >>> L2 = 96 # 96dB
    >>> maad.util.mean_dB(L1,L2)
        93.96292

    Example with arrays

    >>> L1 = [90,80,70]
    >>> maad.util.mean_dB(L1,L1, axis=1)
        array([85.68201724, 85.68201724])
    >>> maad.util.mean_dB(L1,L1, axis=0)  
        array([90., 80., 70.]) 

    """
    # force to be ndarray
    L = np.asarray(argv)

    # test if the 1st dim is = 1 (in case argv is a single 2d array) in order to remove this dimension
    if L.shape[0] == 1:
        L = L[0, ]

    # Verify the adequation between axis number and number of dimensions of L
    if axis >= L.ndim:
        axis = L.ndim - 1

    # dB to energy as sum has to be done with energy
    e = dB2power(L)
    e_mean = e.mean(axis)

    # energy (power) => dB
    e_mean = power2dB(e_mean)

    # test if 2 dimensions but length on 1 dimension is 1
    if e_mean.ndim == 2:
        if (e_mean.shape[0] == 1) or (e_mean.shape[1] == 1):
            e_mean = np.ndarray.flatten(e_mean)

    # test if it's a single value
    if e_mean.size == 1:
        e_mean = float(e_mean)

    return e_mean

# %%


def shift_bit_length(x):
    """
    Find the closest power of 2 that is superior or equal to the number x.

    Parameters
    ----------
    x : scalar

    Returns
    -------
    y : scalar
        the closest power of 2 that is superior or equal to the number x

    Examples
    --------
    >>> maad.util.shift_bit_length(1000)
        1024
    """
    y = 1 << (x-1).bit_length()
    return y

# %%


def nearest_idx(array, value):
    """ 
    Find nearest value on array and return its index.

    Parameters
    ----------
    array: ndarray-like of floats
        array of values where to search the nearest values
    value: float
        value to be searched in array

    Returns
    -------
    idx: int
        index of nearest value on array

    Examples
    --------
    >>> x = np.array([1,2,3])
    >>> maad.util.nearest_idx(x, 1.3)
        0
    >>> maad.util.nearest_idx(x, 1.6)
        1
    """
    # be sure it's ndarray
    array = np.asarray(array)
    idx = (np.abs(array-value)).argmin()
    return idx

# %%


def get_df_single_row(df, index, mode='iloc'):
    """
    Extract a single row from a dataframe keeping the DataFrame type 
    (instead of becoming a Series).

    Parameters
    ----------
    df : pandas DataFrame
        DataFrame with one or more rows

    index : interger (in case of 'iloc') or index (in case of loc)
        index could be the row number or the row index depending on the mode

    mode : string, optional, default is 'iloc'
        choose between row number ('iloc') or row index ('loc')
        (see Pandas documentation for the difference)

    Returns
    -------
    df_out : pandas DataFrame
        Single row DataFrame

    Examples
    --------
    >>> import pandas as pd
    >>> from maad.util import get_df_single_row
    >>> species = [ ('bird', 8, 'Yes' ,'NYC') ,
                  ('insect', 4, 'No','NYC' ) ,
                  ('mammal', 2, 'No','NYC' ) ,
                  ('frog', 3, 'Yes',"LA" ) ]
    >>> df = pd.DataFrame(species, 
                          columns = ['category',
                                     'abundance',
                                     'presence',
                                     'location']) 
    >>> print(df)
          category  abundance presence location
        0     bird          8      Yes      NYC
        1   insect          4       No      NYC
        2   mammal          2       No      NYC
        3     frog          3      Yes       LA
    >>> df.loc[0]
        category     bird
        abundance       8
        presence      Yes
        location      NYC
        Name: 0, dtype: object
    >>> maad.util.get_df_single_row (df, index=0, mode='iloc')
          category abundance presence location
        0     bird         8      Yes      NYC

     Now with index   

    >>> df_modified=df.set_index("category")
    >>> print(df_modified)
                    abundance presence location
        category                             
        bird              8      Yes      NYC
        insect            4       No      NYC
        mammal            2       No      NYC
        frog              3      Yes       LA 
    >>> df_modified.loc['insect']  
        abundance      4
        presence      No
        location     NYC
        Name: insect, dtype: object
    >>> maad.util.get_df_single_row (df_modified, index='insect', mode='loc')
               abundance presence location
        insect         4       No      NYC
    >>> maad.util.get_df_single_row (df_modified, index=1, mode='iloc')    
               abundance presence location
        insect         4       No      NYC
    """

    if mode == 'iloc':
        df_out = pd.DataFrame(df.iloc[index]).T
    elif mode == 'loc':
        df_out = pd.DataFrame(df.loc[index]).T
    else:
        raise TypeError('mode must be iloc or loc, default is iloc')

    return df_out

# %%


def format_features(df, tn, fn):
    """ 
    Format features such as bounding box coordinates and centroids coordinates
    to predifined format : time-frequency to pixels or pixels to time-frequency
    units.

    Parameters
    ----------
    df : pandas DataFrame
        df with bounding box coordinates and/or centroids coordinates       
        array must have a valid input format with column names.

        - bounding box: min_y, min_x, max_y, max_x
        - time frequency: min_f, min_t, max_f, max_t
        - centroid tf : centroid_f, centroid_t
        - centroid pixels : centroid_y, centroid_x

    tn : ndarray
        vector with temporal indices, output from the spectrogram function (in seconds)
    fn: ndarray
        vector with frequencial indices, output from the spectrogram function (in Hz)

    Returns
    -------
    df: pandas DataFrame
        df with bounding box coordinates and/or centroids coordinates in pixels
        and time-frequency units

    Notes
    -----
    Use this function after using functions such as ``maad.rois.select_rois``, 
    ``maad.features.centroid_features``, ``maad.util.read_audacity_annot``, or before using 
    functions such as ``maad.util.overlay_rois``, ``maad.util.overlay_centroid`` in order to 
    format the coordinates into pixels and/or time-frequency units.

    See also
    --------
    select_rois, centroid_features, read_audacity_annot, overlay_rois, 
    overlay_centroid
    
    Examples
    --------
    >>> s, fs = maad.sound.load('../data/cold_forest_daylight.wav')
    >>> Sxx_power, tn, fn, ext = maad.sound.spectrogram(s, fs)
    >>> df_rois = maad.util.read_audacity_annot('../data/cold_forest_daylight_label.txt') 
    >>> list(df_rois)
    >>> df_rois = maad.util.format_features(df_rois, tn, fn)
    >>> list(df_rois)

    """
    # Check format of the input data
    if type(df) is not pd.core.frame.DataFrame:
        raise TypeError('Rois must be of type pandas DataFrame')

    # Do the loop twice in order to force updating coordinates and parameters
    # in pixels unit (time and frequency do not change when the time and 
    # frequency vectors resolution change)
    count = 0
    while (count < 2):

        # if ('min_t' and 'min_f' and 'max_t' and 'max_f') in df and not (('min_y' and 'min_x' and 'max_y' and 'max_x') in df):
        if ('min_t' and 'min_f' and 'max_t' and 'max_f') in df:
            # Check that time and frequency limits are inside tn and fn limits
            # df = df[(df.min_t >= tn.min()) & (df.max_t <= tn.max()) & (df.min_f >= fn.min()) & (df.max_f <= fn.max())]
            if (df.min_t < tn.min()).any() | (df.max_t > tn.max()).any() | (df.min_f < fn.min()).any() | (df.max_f > fn.max()).any():
                warnings.warn(
                    'ROIs boundaries are outside time or frecuency signal limits. Clipping ROIS to valid boundaries.')
                df.loc[df.min_t < tn.min(), 'min_t'] = tn.min()
                df.loc[df.max_t > tn.max(), 'max_t'] = tn.max()
                df.loc[df.min_f < fn.min(), 'min_f'] = fn.min()
                df.loc[df.max_f > fn.max(), 'max_f'] = fn.max()

            bbox = []
            for idx in df.index:
                min_y = nearest_idx(fn, df.loc[idx, 'min_f'])
                min_x = nearest_idx(tn, df.loc[idx, 'min_t'])
                max_y = nearest_idx(fn, df.loc[idx, 'max_f'])
                max_x = nearest_idx(tn, df.loc[idx, 'max_t'])
                # test if max > min, if not, drop row form df
                if (min_y <= max_y) and (min_x <= max_x):
                    bbox.append((min_y, min_x, max_y, max_x))
                else:
                    df = df.drop(idx)
            
            # check if new columns
            if not (('min_y' and 'min_x' and 'max_y' and 'max_x') in df):
                df = df.join(pd.DataFrame(bbox,
                                          columns=['min_y', 'min_x',
                                                   'max_y', 'max_x'],
                                          index=df.index))
            # otherwise update the existing columns with new values
            else:
                df.update(pd.DataFrame(bbox,
                                       columns=['min_y', 'min_x',
                                                'max_y', 'max_x'],
                                       index=df.index))
         

        # if ('min_y' and 'min_x' and 'max_y' and 'max_x') in df and not (('min_t' and 'min_f' and 'max_t' and 'max_f') in df):
        if ('min_y' and 'min_x' and 'max_y' and 'max_x') in df:
            bbox = []
            for _, row in df.iterrows():
                min_f = fn[int(round(row.min_y))]
                min_t = tn[int(round(row.min_x))]
                max_f = fn[int(round(row.max_y))]
                max_t = tn[int(round(row.max_x))]
                # test if max > min, if not, drop row form df
                if (min_f <= max_f) and (min_t <= max_t):
                    bbox.append((min_f, min_t, max_f, max_t))
                else:
                    df = df.drop(row.name)

            # check if new columns
            if not (('min_t' and 'min_f' and 'max_t' and 'max_f') in df):
                df = df.join(pd.DataFrame(bbox,
                                          columns=['min_f', 'min_t',
                                                   'max_f', 'max_t'],
                                          index=df.index))
            

        # if ('centroid_y' and 'centroid_x') in df and not (('centroid_f' and 'centroid_t') in df):
        if ('centroid_y' and 'centroid_x') in df:
            centroid = []
            for _, row in df.iterrows():
                centroid_f = fn[int(round(row.centroid_y))]
                centroid_t = tn[int(round(row.centroid_x))]
                centroid.append((centroid_f, centroid_t))
                
            # check if new columns
            if not (('centroid_f' and 'centroid_t') in df) :
                df = df.join(pd.DataFrame(centroid,
                                          columns=['centroid_f', 'centroid_t'],
                                          index=df.index))

        # if ('centroid_f' and 'centroid_t') in df and not (('centroid_y' and 'centroid_x') in df):
        if ('centroid_f' and 'centroid_t') in df:
            centroid = []
            for idx in df.index:
                centroid_y = nearest_idx(fn, df.loc[idx, 'centroid_f'])
                centroid_x = nearest_idx(tn, df.loc[idx, 'centroid_t'])
                centroid.append((centroid_y, centroid_x))

            # check if new columns
            if not (('centroid_y' and 'centroid_x') in df) :
                df = df.join(pd.DataFrame(centroid,
                                          columns=['centroid_y', 'centroid_x'],
                                          index=df.index))
            # otherwise update the existing columns with new values
            else:
                df.update(pd.DataFrame(centroid,
                                       columns=['centroid_y', 'centroid_x'],
                                       index=df.index))
                
        # =============
        # if ('duration_x' and 'bandwidth_y' and 'area_xy') in df and not (('duration_t' and 'bandwidth_f' and 'area_tf') in df):
        if ('duration_x' and 'bandwidth_y' and 'area_xy') in df:
            area = []
            for _, row in df.iterrows():
                bandwidth_f = row.bandwidth_y * (fn[1]-fn[0])
                duration_t = row.duration_x * (tn[1]-tn[0])
                area_tf = row.area_xy * (fn[1]-fn[0]) * (tn[1]-tn[0])
                area.append((duration_t, bandwidth_f, area_tf))
                
            # check if new columns
            if not (('duration_t' and 'bandwidth_f' and 'area_tf') in df) :
                df = df.join(pd.DataFrame(area,
                                          columns=['duration_t',
                                                    'bandwidth_f', 'area_tf'],
                                          index=df.index))
 
        # if ('duration_t' and 'bandwidth_f' and 'area_tf') in df and not (('duration_x' and 'bandwidth_y' and 'area_xy') in df):
        if ('duration_t' and 'bandwidth_f' and 'area_tf') in df:
            area = []
            for idx in df.index:
                bandwidth_y = round(
                    df.loc[idx, 'bandwidth_f']) / ((fn[1]-fn[0]))
                duration_x = round(df.loc[idx, 'duration_t']) / (tn[1]-tn[0])
                area_xy = round(df.loc[idx, 'area_tf'] /
                                ((fn[1]-fn[0]) * (tn[1]-tn[0])))
                area.append((bandwidth_y, duration_x, area_xy))

            # check if new columns
            if not (('duration_x' and 'bandwidth_y' and 'area_xy') in df):
                df = df.join(pd.DataFrame(area,
                                          columns=['duration_x',
                                                    'bandwidth_y', 'area_xy'],
                                          index=df.index))
            # otherwise update the existing columns with new values
            else:
                df.update(pd.DataFrame(area,
                                       columns=['duration_x',
                                                'bandwidth_y', 'area_xy'],
                                       index=df.index))

        count += 1

    return df


# %%

def crossfade(s1, s2, fs, fade_len=1):
    """
    Add a smooth transition (cross-fade) between two audio signals.
    The function applies a squared function to apply the transition between signals. 
    The signals must have the same sampling rate.
    
    Parameters
    ----------
    s1 : numpy.ndarray
        First audio signal.
    s2 : numpy.ndarray
        Second audio signal.
    fs: int
        Sampling frequency of signals. 

    Returns
    -------
    s_out: numpy.ndarray
        Signal with the resulting crossfade between s1 and s2. The length in seconds 
        of the resulting signal is: `len(s1)/fs + len(s1)/fs - fade_len`. 

    See also
    --------
    crossfade_list
    
    Examples
    --------
    >>> from maad import sound, util
    >>> s, fs = sound.load('../data/spinetail.wav')
    >>> s1 = sound.trim(s, fs, 0, 4)
    >>> s2 = sound.trim(s, fs, 15, 19)
    >>> s_out = util.crossfade(s1, s2, fs, fade_len=1)
    
    """
    fade_in = np.sqrt(np.arange(0, fs * fade_len) / (fs * fade_len))
    fade_out = np.flip(fade_in)
    fstart_idx = len(s1) - (fs * fade_len) - 1
    fstop_idx = fs * fade_len
    s1_fade_out = s1[fstart_idx:-1] * fade_out
    s2_fade_in = s2[0:fstop_idx] * fade_in
    s_out = np.concatenate(
        [s1[0 : fstart_idx + 1], s1_fade_out + s2_fade_in, s2[fstop_idx - 1 : -1]]
    )
    return s_out

# %%


def crossfade_list(s_list, fs, fade_len=1):
    """
    Apply a cross-fade to a list of audio signals.
    Audio signals must be numpy arrays with the same sampling rate. The function applies 
    a squared function to apply the transition between signals.

    Parameters
    ----------
    s_list : list
        List with audio signlas (1D numpy arrays) to be concatenated
    fs : int
        sampling frequency
    fade_len: float
        length of fade between audio in seconds
    
    Returns
    -------
    s_out : ndarray
    
    See also
    --------
    crossfade
    
    """    
    for idx, s in enumerate(s_list):
        if idx == 0:
            s_out = s_list[idx]
        else:
            s = s_list[idx]
            s_out = crossfade(s_out, s, fs)

    return s_out

import os
import urllib.request
import zipfile

def download_audio_dataset(save_location=None, unzip=True):
    url = "https://github.com/juansulloa/audio_dataset/archive/refs/heads/main.zip"
    try:
        # Determine the file name from the URL
        file_name = os.path.basename(url)
        
        # If save_location is None, save to the user's home directory
        if save_location is None:
            home_directory = os.path.expanduser("~")
            save_location = os.path.join(home_directory, 'maad_dataset.zip')
        
        # Download the zip file
        urllib.request.urlretrieve(url, save_location)
        
        print(f"Downloaded and saved {file_name} to {save_location}")
    
        if unzip:
            # Unzip the file if specified
            with zipfile.ZipFile(save_location, 'r') as zipf:
                zipf.extractall(os.path.splitext(save_location)[0])
    
    except Exception as e:
        print(f"An error occurred: {e}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" 
Collection of functions to send queries to www.xeno-canto.org, get dataframe
with all xeno-canto fields and eventually download bird sound files with
JSON metadata.

"""   
#
# Authors:  original author Karoliina Oksanen, 2014
#           Updated to python 3.7.4, Agnieszka Mikolajczyk, 2019
#           Modified for scikit-maad by Sylvain HAUPERT, 2021        
#
# License: New BSD License


import urllib.request
import json
import pandas as pd
from pathlib import Path
import numpy as np
import os

# %%
def xc_query(searchTerms,
             max_nb_files = None,
             format_time = False,
             format_date = False,
             random_seed = 1979,
             verbose=False):
    """
    Query metadata from Xeno-Canto website depending on the search terms. The
    audio recordings metadata are grouped and stored in a dataframe.

    Parameters
    ----------
    searchTerms : list
        list of search terms to perform the query
        The main seach terms are :
        - grp : birds
        - gen : genus
        - ssp : subspecies
        - en  : english name
        - q   : quality
        - cnt : country
        - len : length
        - area : continent (europe, africa, america, asia)
        see more here : https://www.xeno-canto.org/help/search
    max_nb_files: integer, optional
        Maximum number of audio files requested. The default is None
    format_time : boolean, optional
        Time in Xeno-Canto is not always present neither correctly formated. 
        If true, time will be correctly formated to be processed as DateTime 
        format. When formating is not possible, the row is dropped. 
        The default is False
    format_date : boolean, optional
        Date in Xeno-Canto is not always present neither correctly formated. 
        If true, rows with uncorrect format of date are dropped.
    random_seed : integer, optional
        Fix the random seed in order to get the same result every time the 
        function is called
    verbose : boolean, optional
        Print messages during the execution of the function. The default is False.

    Returns
    -------
    df_dataset : pandas DataFrame
        Dataframe containing all the recordings metadata matching search terms
    """
    
    # #*** HACK *** to remove the parameter 'type' from query as it does
    # # not work at the time 10 Nov 2022
    # params = searchTerms
    # searchTerms = []
    # if params is not None :
    #     for param in params:
    #         if 'type' not in param :  
    #             searchTerms.append(param)    
    # #*** END HACK *** 
    
    # initialization of 
    numPages = 1
    page = 1
    df_dataset = pd.DataFrame()
    while page < numPages+1:
        if verbose:
            print("Loading page "+str(page)+"...")
        url = 'https://www.xeno-canto.org/api/2/recordings?query={0}&page={1}'.format(
            '%20'.join(searchTerms), page)
        if verbose:
            print(url)
        jsonPage = urllib.request.urlopen(url)
        jsondata = json.loads(jsonPage.read().decode('utf-8'))
        # check number of pages
        numPages = jsondata['numPages']
        # Append pandas dataframe of records & convert to .csv file
        df_dataset = pd.concat([df_dataset, pd.DataFrame(jsondata['recordings'])]) #df_dataset.append(pd.DataFrame(jsondata['recordings']))
        # increment the current page
        page = page+1
                
    # test if the dataset is not empty
    if len(df_dataset)>0:
        
        # #*** HACK *** to filter the dataset with the parameter 'type' as it does
        # # not work for regular query at the time 10 Nov 2022
        # if verbose :
        #     print("searchTerms {}".format(searchTerms))
        # if params is not None :
        #     for param in params  :
        #         if 'type' in param :
        #             value =  param.split(':')[1] 
        #             df_dataset = df_dataset[df_dataset.type.apply(lambda type: value in type)]
        # #*** END HACK *** 

        # convert latitude and longitude coordinates into float
        df_dataset['lat'] = df_dataset['lat'].astype(float)
        df_dataset['lng'] = df_dataset['lng'].astype(float)
        
        # rearrange index to be sure to have unique and increasing index
        df_dataset.reset_index(drop=True, inplace=True)
        
        # the format of length is not correct (missing 0 before 0:45 => 00:45)
        # Correct the format of length for length shorten than 9:59 (4 characters)
        # by adding a 0
        df_dataset['length'].where(~(df_dataset.length.str.len()==4), 
                                    other='0'+ df_dataset[df_dataset.length.str.len()==4].length, 
                                    inplace=True) 
        
        if format_time == True :
            # rearrange index to be sure to have unique and increasing index
            df_dataset.reset_index(drop=True, inplace=True)
            
            # the format of time is not always correct
            # replace . by :
            df_dataset['time'].replace(to_replace = '[.]', value=':', regex= True)
            df_dataset['time'].replace(to_replace = '[ ] ', value='', regex= True)
            
            # drop rows where there is no valid time information that can be corrected
            df_dataset = df_dataset[(df_dataset.time.str.match('^(0[0-9]|1[0-9]|2[0-3])[:]([0-5][0-9])$')) | 
                                    (df_dataset.time.str.match('^([0-9])[:]([0-5][0-9])$'))]
            
            # Correct the format of time when 0 is missing (missing 0 before 0:45 => 00:45)
            # by adding a 0
            df_dataset['time'][df_dataset.time.str.match('^([0-9])[:]([0-5][0-9])$')] = '0' + df_dataset[df_dataset.time.str.match('^([0-9])[:]([0-5][0-9])$')].time

            if verbose:
                print("Keeped metadata for", len(df_dataset), "files after formating time")
            
        if format_date == True :
            # rearrange index to be sure to have unique and increasing index
            df_dataset.reset_index(drop=True, inplace=True)
            # drop rows where there is no valid date information
            df_dataset = df_dataset[df_dataset.date.str.match(r'^(20[0-9][0-9]|19[0-9][0-9])-(0[1-9]|1[0-2])-([1-9]|1[0-9]|2[0-9]|3[0-1])$')]
            
            if verbose:
                print("Keeped metadata for", len(df_dataset), "files after formating date")
        
        if (format_time == True) and (format_date == True) :         
            # add a column with the week number
            df_dataset['week'] = pd.to_datetime(df_dataset['date']).dt.isocalendar()['week'] # type: ignore
            # add a column with datetime in DateTime format
            df_dataset['datetime'] =  pd.to_datetime(df_dataset['time']+' '+ df_dataset['date'], format="%H:%M %Y-%m-%d")

    # if no limit in the number of files
    if max_nb_files is not None :
        # test if the number of files is greater than the maximum number of
        # resquested files
        if len(df_dataset) > max_nb_files : 
            df_dataset = df_dataset.sample(n = max_nb_files,
                                           random_state = random_seed)
       
    if verbose:
        print("Found", numPages, "pages in total.")
        print("Saved metadata for", len(df_dataset), "files")
            
    # rearrange index to be sure to have unique and increasing index
    df_dataset.reset_index(drop=True, inplace=True)

    return df_dataset

# %%
def xc_multi_query(df_query,
                   max_nb_files = None,
                   format_time  = False,
                   format_date  = False,
                   random_seed  = 1979,
                   verbose      = False):
    """
    Multi_query performs multiple queries following the search terms defined
    in the input dataframe

    Parameters
    ----------
    df_query : pandas DataFrame
        Dataframe with search terms. Each row corresponds to a new query. 
        Columns corresponds to the search terms allowed by Xeno-Canto
    max_nb_files: integer, optional
        Maximum number of audio files requested. The default is None
    format_time : boolean, optional
        Time in Xeno-Canto is not always present neither correctly formated. 
        If true, time will be correctly formated to be processed as DateTime 
        format. When formating is not possible, the row is dropped. 
        The default is False
    format_date : boolean, optional
        Date in Xeno-Canto is not always present neither correctly formated. 
        If true, rows with uncorrect format of date are dropped.
    random_seed : integer, optional
        Fix the random seed in order to get the same result every time the 
        function is called
    verbose : boolean, optional
        Print messages during the execution of the function. The default is False.

    Returns
    -------
    df_dataset : pandas DataFrame
        Dataframe containing all the recordings metadata matching 
        the search terms.

    """

    df_dataset = pd.DataFrame()
    for index, row in df_query.iterrows():
        searchTerms = row.tolist()
        df_dataset = pd.concat([df_dataset, xc_query(searchTerms, 
                                                max_nb_files,
                                                format_time,
                                                format_date,
                                                random_seed,
                                                verbose)]) # df_dataset.append

    # rearrange index to be sure to have unique and increasing index
    df_dataset.reset_index(drop=True, inplace=True)
    
    return df_dataset

#%%
def xc_selection(df_dataset,
                 max_nb_files=100,
                 max_length='01:00',
                 min_length='00:10',
                 min_quality='B',
                 verbose = False):
    """
    Select a maximum number of recordings depending on their quality and 
    duration in order to create an homogeneous dataset.

    Parameters
    ----------
    df_dataset : pandas DataFrame
        Dataframe containing all the recordings metadata 
    max_nb_files : int, optional
        Max number of audio files per species. The default is 100.
    max_length : string, optional
        Max duration of the audio files. The default is '01:00'.
    min_length : string, optional
        Min duration of the audio files. The default is '00:10'.
    min_quality : string, optional
        Min quality of the audio files. The default is 'B'.
    verbose : boolean, optional
        Print messages during the execution of the function. The default is False.

    Returns
    -------
    df_dataset_out : pandas DataFrame
        Dataframe containing the selected recordings metadata 

    """

    df_dataset_out = pd.DataFrame()
    quality = ['A', 'B', 'C', 'D', 'E']
    if min_quality == 'A' :
        quality =['A']
    elif min_quality == 'B' :
        quality = ['A', 'B']
    elif min_quality == 'C':
        quality = ['A', 'B', 'C']
    elif min_quality == 'D':
        quality = ['A', 'B', 'C', 'D']
    unique_species = pd.unique(df_dataset.gen + ' ' + df_dataset.sp)
    for name in unique_species:
        if verbose : print(name)
        # extract the genus and species from the scientific name
        gen = name.rpartition(' ')[0]
        sp = name.rpartition(' ')[2]
        # select the corresponding to the species
        # !! the string test is case sensitive (the genus start with a upper case)
        subdf_dataset = df_dataset[((df_dataset.gen == gen) &
                                    (df_dataset.sp == sp))]
        # sort the dataframe corresponding to the species by audio quality
        subdf_dataset = subdf_dataset.sort_values(by='q')

        # Counter initialization
        current_nb_files = 0
        requested_nb_files = 0
        current_quality = 0
        while (current_nb_files < max_nb_files) & (current_quality < len(quality)):
            requested_nb_files = max_nb_files - current_nb_files
            q = quality[current_quality]
            if verbose : print('    ... request %2.0f files of quality %s' %
                  (requested_nb_files, q))
            
            mask1 = ((subdf_dataset.q == q) &
                    (subdf_dataset.length <= max_length) &
                    (subdf_dataset.length >= min_length))

            if len(subdf_dataset[mask1]) >= requested_nb_files:
                # create a temp dataframe with the selected rows
                df_temp = subdf_dataset[mask1].sort_values(
                    by='length', ascending=False).iloc[0:requested_nb_files]
                # add the rows to the output dataframe
                df_dataset_out = pd.concat([df_dataset_out, df_temp], axis=0)
                # drop the selected rows to avoid future selection
                subdf_dataset.drop(df_temp.index, axis=0, inplace=True)
            else :  
                # create a temp dataframe with the selected rows
                df_temp = subdf_dataset[mask1].sort_values(
                    by='length', ascending=False)
                # add the rows to the output dataframe
                df_dataset_out = pd.concat([df_dataset_out, df_temp], axis=0)
                # drop the selected rows to avoid future selection
                subdf_dataset.drop(df_temp.index, axis=0, inplace=True)
            if verbose : 
                print('    --> found %2.0f files of quality %s and %s<length<%s'%
                     (len(df_temp), q, min_length, max_length ))
            current_nb_files += len(df_temp)
            requested_nb_files = max_nb_files - current_nb_files
            current_quality += 1

        if verbose : print('    total files : %2.f' %current_nb_files)
        if verbose : print("-----------------------------------------")
        
    return df_dataset_out

#%%
def xc_download(df,
                rootdir,
                dataset_name='dataset',
                overwrite   = False,
                save_csv    = False,
                verbose     = False):
    """
    Download the audio files from Xeno-Canto based on the input dataframe
    It will create directories for each species if needed

    Parameters
    ----------
    df : pandas DataFrame
        Dataframe containing the selected recordings metadata 
    rootdir : string
        Path to the directory where the whole dataset will be saved
    dataset_name : string, optional
        Name of the dataset that will be created as a parent directory . 
        The default is 'dataset'.
    overwrite : boolean, optional
        Test if the directory where the audio files will be downloaded already
        exists. if True, it will download the data in the directory anyway.
        Otherwise, if False, it will not download audio files.
    save_csv : boolean, optional
        if True, the csv corresponding to the species will be saved in the
        directory of the species. The default is False.
    verbose : boolean, optional
        Print messages during the execution of the function. The default is False.

    Returns
    -------
    df : pandas DataFrame
        Dataframe similar to df but without the rows of the audio recordings
        that were not downloaded.
        Add a new column "fullfilename" with the paths to the newly downloaded 
        audio files
    """
    # format rootdir as path
    rootdir = Path(rootdir)

    # list of the full paths to the audios
    fullpath_list = []
    
    # Try to set 'id' as index
    try :
        df.set_index('id', inplace = True)
    except :
        pass
   
    #--------------------------------------------------------------------------
    # Check whether the specified path is an existing directory or not 
    isdir = os.path.exists(rootdir / dataset_name)    
    if (isdir == False) or ((isdir == True) and (overwrite == True)) : 
        if (overwrite == True):
            if verbose:
                print(
                    "The directory "
                    + str(rootdir / dataset_name)
                    + " already exists and will be overwritten" )        
        if verbose :
            numfiles = len(df)
            print("A total of", numfiles, "files will be downloaded")
            
        # change type of rootdir into Path
        count = 1
        for index, row in df.iterrows():

            #------------------------------------------------------------------
            # create a name for the directory
            name_dir = row.gen + ' ' + row.sp + '_' + row.en
            # create a directory for the species
            path = rootdir / dataset_name / name_dir
            if not os.path.exists(path):
                if verbose :
                    print("Creating subdirectory " + 
                          str(path) + 
                          " for downloaded files...")
                os.makedirs(path)
            #------------------------------------------------------------------
            # get filenames
            audio_format = '.' + row['file-name'].split('.')[-1]
            filename = 'XC' + str(index) + audio_format
            # test if the file already exists
            if os.path.exists(rootdir / dataset_name / name_dir / filename) == True:
                fullpath_list += [path / filename]
                if verbose :
                    print( filename + " already exists")
                # drop the row of this recordings
                #df.drop(index, inplace = True)
            else:
                #--------------------------------------------------------------
                # get website recording http download address 
                fileaddress = row.file
                # try to download the audio recording
                try :
                    fullpath, _ = urllib.request.urlretrieve(fileaddress, path / filename)
                    fullpath_list += [str(fullpath)]
                    if verbose : 
                        numfiles = len(df)
                        print("Saving file ", count, "/", numfiles, ": " + fileaddress)
                except:
                    # can't download the audio file (it does not exist (anymore) in
                    # xeno-canto)
                    if verbose :
                        numfiles = len(df)
                        print("***WARNING*** Can't save the file ", 
                              count, "/", numfiles, ": " + fileaddress)
                    # drop the row of this recordings
                    df.drop(index, inplace = True)
                    
            #------------------------------------------------------------------        
            # save csv
            if save_csv:
                filename_csv = str(path/'metadata.csv')
                # test if the csv file doesn't exit
                if os.path.exists(filename_csv) == False:
                    # try to create a file and add a row corresponding to the index
                    try :
                        df.loc[index].to_frame().T.to_csv(filename_csv, 
                                                          sep=";", 
                                                          index=True, 
                                                          index_label = 'id') 
                    except :
                        pass
                # if the csv file exists, concat both dataframes
                else :
                    # try to read the file and add a row corresponding to the index
                    try :
                        pd.concat([pd.read_csv(filename_csv,sep=';',index_col='id'), 
                                   df.loc[index].to_frame().T], 
                                  ignore_index=False).drop_duplicates().to_csv(filename_csv, 
                                                                               sep=";", 
                                                                               index=True,
                                                                               index_label='id')       
                    except :
                        pass
            
            # increment the counter
            count += 1
                        
        # add a new column
        df['fullfilename'] = fullpath_list
        
    else :
        if verbose:
            print(
                "***WARNING*** : The directory "
                + str(rootdir)
                + " already exists"
            )
            
    return df

#%%

# def xc_save_csv(
#         df,
#         rootdir   = os.getcwd(),
#         filename  = "xc_metadata.csv",
#         overwrite = False,
#         verbose   = False): 
#     """
#     Save audio recordings metadata collected from xeno-canto into a csv file

#     Parameters
#     ----------
#     df : pandas DataFrame
#         Dataframe containing the selected recordings metadata 
#     rootdir : string, optional
#         Path to the directory. The default is the current directory
#     filename : string, optional
#         Name of the csv file. The default is "xc_metadata.csv"
#     overwrite : boolean, optional
#         Overwirte the csv file if it already exists
#     verbose : boolean, optional
#         Print messages during the execution of the function. The default is False.

#     Returns
#     -------
#     fullpath : string
#         Returns the full path of the csv file
#     """    
#     # format rootdir to Path
#     rootdir = Path(rootdir)
    
#     # Check whether the specified path is an existing file or not 
#     isfile = os.path.isfile(rootdir / filename)
    
#     if (isfile == False) or ((isfile == True) and (overwrite == True)) : 
#         if (overwrite == True):
#             if verbose:
#                 print(
#                     "The file "
#                     + filename
#                     + " already exists in " 
#                     + str(rootdir)
#                     + " and will be overwritten" )      
#         df.to_csv(rootdir / filename, 
#                   sep=';', 
#                   index=True, 
#                   header=True)
#         fullpath = rootdir / filename
#     else:
#         if verbose:
#             print(
#                 "***WARNING*** : The file "
#                 + filename
#                 + " already exists in "
#                 + str(rootdir)
#             )
#         fullpath = []
        
#     return fullpath
    
# #%%
    
# def xc_read_csv(        
#         filename,
#         rootdir   = os.getcwd(),
#         verbose   = False): 
#     """
#     Read audio recordings metadata collected from xeno-canto and saved into
#     a csv file

#     Parameters
#     ----------
#     filename : string
#         Name of the csv file.
#     rootdir : string, optional
#         Path to the directory. The default is the current directory
#     verbose : boolean, optional
#         Print messages during the execution of the function. The default is False.

#     Returns
#     -------
#     df_dataset : pandas DataFrame
#         Dataframe containing the audio recordings metadata 
#     """    
#     # format rootdir to Path
#     rootdir = Path(rootdir)
    
#     try :
#         df_dataset = pd.read_csv(rootdir / filename, sep=';', index='id')
#     except:
#         df_dataset = pd.DataFrame()
#         if verbose : 
#             print(
#                 "***WARNING : The file "
#                 + filename
#                 + " does not exist in " 
#                 + str(rootdir))
    
#     return df_dataset
    
# %%
if __name__ == '__main__':
    
    df_query = pd.DataFrame()
    df_species = pd.DataFrame()
    
    # species
    df_species['scientific name'] = ['Agelaius phoeniceus',
                                     'psittacula krameri',
                                     'Ardea herodias']
    # query
    gen = []
    sp = []
    for name in df_species['scientific name']:
        gen.append(name.rpartition(' ')[0])
        sp.append(name.rpartition(' ')[2])

    df_query['gen'] = gen
    df_query['sp'] = sp
    df_query['q'] = 'q:A'
    df_query['len'] = 'len:"10-60"'

    df_dataset = xc_multi_query(df_query, 
                                max_nb_files = 5,
                                verbose=True)
    
    print('number of files in the dataset is %d'%(len(df_dataset)))
    print(df_dataset.head(15))
    
    
    xc_download(df_dataset, 
                rootdir=os.getcwd(), 
                dataset_name='my_dataset', 
                save_csv=True,
                overwrite=False,
                verbose = True)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Template matching functions

New BSD License

TODO:
    - min_t and max_t should be 0 to len(Sxx_audio)
"""

import pandas as pd
import numpy as np
from scipy.signal import find_peaks
from skimage.feature import match_template
import matplotlib.pyplot as plt
from matplotlib import patches
from maad import sound, util

#%%
def template_matching(
    Sxx, Sxx_template, tn, ext, peak_th, peak_distance=None, display=False, **kwargs
):
    """
    Use normalized spectrogram cross-correlation to detect the occurrence of a template
    sound in a target audio.

    The output is an array with values between -1.0 and 1.0. The value at a given
    position corresponds to the correlation coefficient between the spectrograms of
    the target audio and the template. The function also delivers the detection found
    as peaks and as regions of interest (rois).

    Parameters
    ----------
    Sxx : 2D array
        Spectrogram of audio signal.

    Sxx_template : TYPE
        Spectrogram of target sound.

    tn : 1d array
        Time vector of target audio, which results from the maad.sound.spectrogram function.

    fn : 1d array
        Frecuency vector of target audio, which results from the maad.sound.spectrogram function.

    ext : list of scalars [left, right, bottom, top]
        Extent keyword arguments controls the bounding box in data coordinates for the
        spectrogram of the target audio, which results from the maad.sound.spectrogram function.

    peak_th : float, optional
        Threshold applied to find peaks in the cross-correlation array.
        Should be a value between -1 and 1.

    peak_distance : float, optional
        Required minimal temporal distance (>= 0) in seconds between neighbouring
        peaks. If set to `None`, the minimum temporal resolution will be used.
        The minimal temporal resolution is given by the array tn and depends on the parameters
        used to compute the spectrogram.

    display : Boolean, optional
        display the results of the template matching. The default is False.

    **kwargs: keywords pair, optional
        Set aditional specificities to find peaks in xcorroelation array.
        Arguments are passed to the the function scipy.signal.find_peaks.


    Returns
    -------
    xcorrcoef : 1D array
        Correlation coefficients resulting from the cross-correlation between
        audio and target template.

    rois : pandas DataFrame
        Detections found based on cross-correlation coefficients. The result is
        presented as a DataFrame where each row represent a detection, with the
        peak time (peak_time), peak amplitude (xcorrcoef), minimum and maximum time
        (min_t, max_t), and minimum and maximum frequency (min_f, max_f).

    """

    # check inputs
    if Sxx.ndim < Sxx.ndim:
        raise ValueError(
            "Dimensionality of template must be less than or "
            "equal to the dimensionality of image."
        )
    if np.any(np.less(Sxx.shape, Sxx_template.shape)):
        raise ValueError("Target spectrogram must be larger than template.")

    if peak_distance is None:  # if not provided, set to minimum distance
        peak_distance = np.diff(tn)[0]

    # set temporal distance to spectrogram pixels
    peak_distance_pixel = peak_distance / np.diff(tn)[0]

    if peak_distance_pixel < 1:
        raise ValueError(
            f"`peak_distance` must be greater or equal to spectrogram resolution: {np.diff(tn)[0]}"
        )

    # Pad Sxx to have len(xcorrcoef) == Sxx.shape[1]
    # if Sxx_template.shape[1] is even substract 1 to time width
    time_width = np.floor(Sxx_template.shape[1] / 2).astype(int)
    if (Sxx_template.shape[1] % 2) == 1:
        pad_width = (
            (
                0,
                0,
            ),
            (time_width, time_width),
        )
    else:
        pad_width = (
            (
                0,
                0,
            ),
            (time_width, time_width - 1),
        )
    Sxx_pad = np.pad(Sxx, pad_width, mode="edge")

    # Compute normalized cross-correlation
    xcorrcoef = match_template(Sxx_pad, Sxx_template)

    # When flims from Sxx is larger than Sxx_template, take mean value
    xcorrcoef = np.mean(xcorrcoef, axis=0)

    ## Find peaks
    prominence = kwargs.pop("prominence", None)
    width = kwargs.pop("width", None)
    wlen = kwargs.pop("wlen", None)
    rel_height = kwargs.pop("rel_height", 0.5)
    plateau_size = kwargs.pop("plateau_size", None)
    threshold = kwargs.pop("threshold", None)

    peaks, peak_dict = find_peaks(
        xcorrcoef,
        peak_th,
        threshold,
        peak_distance_pixel,
        prominence,
        width,
        wlen,
        rel_height,
        plateau_size,
    )
    peaks_time = tn[peaks]

    # Build rois as pandas Dataframe
    # Create Dataframe and adjust extreme values for min_t and max_t
    template_len = tn[Sxx_template.shape[1]] - tn[0]
    rois = pd.DataFrame(
        {
            "peak_time": peaks_time,
            "xcorrcoef": xcorrcoef[peaks],
            "min_t": peaks_time - template_len / 2,
            "max_t": peaks_time + template_len / 2,
        }
    )
    rois.loc[rois.min_t < 0 , "min_t"] = tn[0]
    rois.loc[rois.max_t > tn[-1] , "max_t"] = tn[-1]

    if display == True:
        rois['min_f'] = ext[2]
        rois['max_f'] = ext[3]
        # plot spectrogram
        fig, ax = plt.subplots(2, 1, figsize=(8, 5), sharex=True)
        util.plot_spectrogram(Sxx, ext, log_scale=False, ax=ax[0], colorbar=False)
        if not (rois.empty):
            for idx, _ in rois.iterrows():
                xy = (rois.min_t[idx], rois.min_f[idx])
                width = rois.max_t[idx] - rois.min_t[idx]
                height = rois.max_f[idx] - rois.min_f[idx]
                rect = patches.Rectangle(
                    xy, width, height, lw=1, edgecolor="yellow", facecolor="yellow", alpha=0.25
                )
                ax[0].add_patch(rect)

        # plot corr coef
        ax[1].plot(tn[0 : xcorrcoef.shape[0]], xcorrcoef)
        ax[1].plot(peaks_time, xcorrcoef[peaks], "x")
        ax[1].hlines(peak_th, 0, tn[-1], linestyle="dotted", color="0.75")
        ax[1].set_xlabel("Time [s]")
        ax[1].set_ylabel("Correlation coeficient")

    return xcorrcoef, rois



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" 
Collection of functions to estimation sound attenuation due to spreading loss,
atmospheric absorption and habitat attenuation (diffraction, reflexion, 
diffusion... due to folliage, ground, trunks).These functions are required to
estimate active distance (i.e. detection range) of audio recorder (such as SM4)
 as well as initial sound pressure level (dB SPL).
"""   
#
# Authors:  Juan Sebastian ULLOA <lisofomia@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>        
#
# License: New BSD License

#%%
#***************************************************************************
# -------------------       Load modules         ---------------------------
#***************************************************************************
# Import external modules
import numpy as np 
from numpy import sum, log10, sqrt, exp
import pandas as pd

# min value
import sys
_MIN_ = sys.float_info.min

# import internal modules
from maad.spl import dBSPL2pressure, pressure2dBSPL

#%%
# =============================================================================
# Private functions
# =============================================================================

def _geometric_att_factor (r,r0) :
  """
  Get the attenuation factor due to spreading loss (also known as geometric
  attenuation). Usually the source is considered to be ponctual which creates
  a spherical spreading loss.
  
  Parameters
  ----------
  r : scalar or array-like
      propagation distances in m [SCALAR or VECTOR]
  r0 : scalar
      reference distance in m [SCALAR]
  
  Returns
  -------
  Ageo_factor : scalar or array-like
      Geometric attenuation factor 
      => Multiply this factor with the effective reference acoustic pressure p0 
      measured at r0 to estimate the pressure after attenuation

  """ 
  # make sure it's array
  r = np.asarray(r)  
  
  Ageo_factor = r0/r
  
  return Ageo_factor

#%%
def _geometric_att_dB (r,r0) :
  """
  Get the attenuation in dB due to spreading loss (also known as geometric 
  attenuation).Usually the source is considered to be ponctual  which creates 
  a spherical spreading loss.
  
  Parameters
  ----------
  r : scalar or array-like
      propagation distances in m 
  r0 : scalar
      reference distance in m
  
  Returns
  -------
  Ageo_dB : scalar or array-like
      Geometric attenuation 
      => Subtract this value from the reference acoustic pressure L0 in dB 
      (or sound pressure level (SPL)) measured at r0 to estimate the
      acoustic pressure L in dB after attenuation
      
  """ 
  # make sure it's array
  r = np.asarray(r)    
  
  Ageo_dB = -20*log10(_geometric_att_factor(r,r0))
  
  return Ageo_dB

#%%
def _atmospheric_att_coef_dB (f, t=20, rh=60, pa=101325):
   
  """ 
  Get the atmospheric attenuation coefficient in dB/m  
  
  Parameters
  ----------
  f: scalar or array-like
      frequency in Hz
  t: scalar, optional, default is 20
      temperature in °C
  rh: scalar, optional, default is 60
      relative humidity in % 
  pa: scalar, optional, default is 101325
      atmospheric pressure in Pa 
  
  Returns
  -------
  Aatm_coef_dB : scalar or array-like
      atmospheric attenuation coefficient in dB/m  
  
  References
  ----------
  Partially from http://www.sengpielaudio.com/AirdampingFormula.htm
  """
  # make sure it's array
  f = np.asarray(f)  
  # test if t, rh and pa are scalars
  if hasattr(t, "__len__") : raise TypeError ('t must be a scalar ')    
  elif hasattr(rh, "__len__") : raise TypeError ('rh must be a scalar ')    
  elif  hasattr(pa, "__len__") : raise TypeError('pa must be a scalar ')       
  
  pr = 101.325e3 # reference ambient atmospheric pressure: 101.325 kPa
  To1 = 273.16 # triple-point isotherm temp: 273.16 K
  To = 293.15 #  reference temperature in K: 293.15 K (20°C)
  t = t+273.15 # celcius to farenheit
  
  psat = pr*10**(-6.8346 * (To1/t)**1.261 + 4.6151) #saturation vapor pressure equals
  h = rh * (psat / pa) # molar concentration of water vapor, as a percentage
  frO = (pa / pr) * (24 + 4.04e4 * h * ((0.02 + h) / (0.391 + h))) # oxygen relaxation frequency
  frN = (pa / pr) * sqrt(t / To) * (9 + 280 * h * exp(-4.170*((t/To)**(-1/3) -1))) # nitrogen relaxation frequency
  
  z = 0.1068 * exp (-3352/t) / (frN+f**2 /frN)
  y = (t/To)**(-5/2) * (0.01275 * exp(-2239.1/t) * 1/(frO+f**2/frO) + z)
  Aatm_coef_dB = 8.686 * f**2 * ((1.84e-11 * 1/(pa/pr) * sqrt(t/To)) + y)
  
  return Aatm_coef_dB 

#%%
def _atmospheric_att_coef (f, t=20, rh=60, pa=101325):
  """ 
  Get the atmospheric attenuation coefficient in Neper/m  
  
  Parameters
  ----------
  f: scalar or array-like
      frequency in Hz
  t: scalar, optional, default is 20
      temperature in °C
  rh: scalar, optional, default is 60
      relative humidity in % 
  pa: scalar, optional, default is 101325
      atmospheric pressure in Pa 
  
  Returns
  -------
  Aatm_coef: scalar or array-like
      atmospheric attenuation coefficient in Neper/m  
  
  References
  ----------
  Partially from http://www.sengpielaudio.com/AirdampingFormula.htm
  """  
  # make sure it's array
  f = np.asarray(f)  
  # test if t, rh and pa are scalars
  if hasattr(t, "__len__") : raise TypeError ('t must be a scalar ')    
  elif hasattr(rh, "__len__") : raise TypeError ('rh must be a scalar ')    
  elif  hasattr(pa, "__len__") : raise TypeError('pa must be a scalar ')  
    
  Aatm_coef = _atmospheric_att_coef_dB (f, t, rh, pa)/(20*log10(exp(1)))
  
  return Aatm_coef 
    
def _atmospheric_att_factor (f, r, r0, t=20, rh=60, pa=101325):

  """ 
  Get the atmospheric attenuation factor 
  
  Parameters
  ----------
  f: scalar or array-like
      frequency in Hz
  r : scalar or array-like
      propagation distances in m 
  r0 : scalar
      reference distance in m
  t: scalar, optional, default is 20
      temperature in °C
  rh: scalar, optional, default is 60
      relative humidity in % 
  pa: scalar, optional, default is 101325
      atmospheric pressure in Pa 
  
  Returns
  -------
  Aatm_factor : scalar or array-like (vector (1D) or matrix (2D))
      atmospheric attenuation factor 
      => Multiply this factor with the effective reference acoustic pressure p0 
      measured at r0 to estimate the pressure after attenuation
      rows : frequencies
      columns :distances
  """
  # make sure it's array
  r = np.asarray(r)
  f = np.asarray(f) 
  
  # test if r is a single value
  if r.ndim == 0 : Nr= 1
  else: Nr = len(r.flatten())
  
  # test if f is a single value
  if f.ndim == 0 : Nf= 1
  else: Nf = len(f.flatten())
  
  # test if t, rh and pa are scalars
  if hasattr(t, "__len__") : raise TypeError ('t must be a scalar ')    
  elif hasattr(rh, "__len__") : raise TypeError ('rh must be a scalar ')    
  elif  hasattr(pa, "__len__") : raise TypeError('pa must be a scalar ')   

  Aatm_coef = _atmospheric_att_coef(f, t, rh, pa)
  Aatm_factor = exp(-Aatm_coef.reshape(Nf,1) @ (r.reshape(1,Nr)-r0))
  
  # reshape the array when dimensions are 1
  if Aatm_factor.shape[0] == 1:
      Aatm_factor = Aatm_factor[0][:]
  elif Aatm_factor.shape[1] == 1:
      Aatm_factor = Aatm_factor[:,0]
  
  return Aatm_factor

def _atmospheric_att_dB (f, r, r0, t=20, rh=60, pa=101325):

  """ 
  Parameters
  ----------
  f: scalar or array-like
      frequency in Hz
  r : scalar or array-like
      propagation distances in m 
  r0 : scalar
      reference distance in m
  t: scalar, optional, default is 20
      temperature in °C
  rh: scalar, optional, default is 60
      relative humidity in % 
  pa: scalar, optional, default is 101325
      atmospheric pressure in Pa 
  
  Returns
  -------
  Aatm_dB : scalar or array-like (vector (1D) or matrix (2D))
      atmospheric attenuation in dB depending on the frequency, the temperature,
      the atmospheric pressure, the relative humidity and the distance 
      => subtract Aatm_dB from the reference acoustic pressure L0 in dB 
      (or sound pressure level (SPL)) measuread at distance r0 for each 
      frequency and distance
      rows : frequencies
      columns :distances
  """
  # make sure it's array
  r = np.asarray(r)
  f = np.asarray(f) 
  
  # test if r is a single value
  if r.ndim == 0 : Nr= 1
  else: Nr = len(r.flatten())
  
  # test if f is a single value
  if f.ndim == 0 : Nf= 1
  else: Nf = len(f.flatten())
  
  # test if r0, t, rh and pa are scalars
  if hasattr(r0, "__len__") : raise TypeError ('r0 must be a scalar ')   
  elif hasattr(t, "__len__") : raise TypeError ('t must be a scalar ')    
  elif hasattr(rh, "__len__") : raise TypeError ('rh must be a scalar ')    
  elif  hasattr(pa, "__len__") : raise TypeError('pa must be a scalar ') 
  
  Aatm_coef_dB = _atmospheric_att_coef_dB(f, t, rh, pa)
  Aatm_dB = Aatm_coef_dB.reshape(Nf,1) @ (r.reshape(1,Nr)-r0)
  
  # reshape the array when dimensions are 1
  if Aatm_dB.shape[0] == 1:
      Aatm_dB = Aatm_dB[0][:]
  elif Aatm_dB.shape[1] == 1:
      Aatm_dB = Aatm_dB[:,0]

  return Aatm_dB

#%%
def _habitat_att_factor (f, r, r0, a0=0.02):
  """ 
  Get the habitat attenuation factor
  
  Parameters
  ----------
  f: scalar or array-like
      frequency in Hz
  r : scalar or array-like
      propagation distances in m 
  r0 : scalar
      reference distance in m
  a0 : scalar
      attenuation coefficient of the habitat in dB/kHz/m 
  
  Returns
  -------
  Ahab : scalar or array-like (vector (1D) or matrix (2D))
      habitat attenuation depending on the frequency and the distance [MATRIX] 
      => Multiply this value with the effective reference acoustic pressure p0 
      measured at r0 to estimate the pressure after attenuation
  """
  # make sure it's array
  r = np.asarray(r)
  f = np.asarray(f) 

  # test if r is a single value
  if r.ndim == 0 : Nr= 1
  else: Nr = len(r.flatten())
  
  # test if f is a single value
  if f.ndim == 0 : Nf= 1
  else: Nf = len(f.flatten())
  
  # test if r0 and a0 are scalars
  if hasattr(r0, "__len__") : raise TypeError ('r0 must be a scalar ')   
  elif hasattr(a0, "__len__") : raise TypeError ('a0 must be a scalar ')    

  Ahab_coef = _habitat_att_coeff(f, a0)
  Ahab_factor = exp(-Ahab_coef.reshape(Nf,1) @ (r.reshape(1,Nr)-r0))
  
  # reshape the array when dimensions are 1
  if Ahab_factor.shape[0] == 1:
      Ahab_factor = Ahab_factor[0][:]
  elif Ahab_factor.shape[1] == 1:
      Ahab_factor = Ahab_factor[:,0]
  
  return Ahab_factor

#%%
def _habitat_att_dB(f, r, r0, a0=0.02) :
  """ 
  Get the habitat attenuation in dB
  
  Parameters
  ----------
  f: scalar or array-like
      frequency in Hz
  r : scalar or array-like
      propagation distances in m 
  r0 : scalar
      reference distance in m
  a0 : scalar, optional, default is 0.02
      attenuation coefficient of the habitat in dB/kHz/m 
    
  Returns
  -------
  Ahab_dB : scalar or array-like (vector (1D) or matrix (2D))
      habitat attenuation in dB depending on the frequency, the distance and 
      the habitat attenuation coefficient
      => subtract Aatm_dB from the reference acoustic pressure L0 in dB 
      (or sound pressure level (SPL)) measuread at distance r0 for each 
      frequency and distance
      rows : frequencies
      columns :distances
  """  
  #Ahab_dB = -20*log10(_habitat_att_factor(f,r,r0,a0))
  
  # make sure it's array
  r = np.asarray(r)
  f = np.asarray(f) 

  # test if r is a single value
  if r.ndim == 0 : Nr= 1
  else: Nr = len(r.flatten())
  
  # test if f is a single value
  if f.ndim == 0 : Nf= 1
  else: Nf = len(f.flatten())
  
  # test if r0 and a0 are scalars
  if hasattr(r0, "__len__") : raise TypeError ('r0 must be a scalar ')   
  elif hasattr(a0, "__len__") : raise TypeError ('a0 must be a scalar ')   
  
  Ahab_coef_dB = _habitat_att_coeff_dB(f,a0)
  Ahab_dB = Ahab_coef_dB.reshape(Nf,1) @ (r.reshape(1,Nr)-r0)
  
  # reshape the array when dimensions are 1
  if Ahab_dB.shape[0] == 1:
      Ahab_dB = Ahab_dB[0][:]
  elif Ahab_dB.shape[1] == 1:
      Ahab_dB = Ahab_dB[:,0]
    
  return (Ahab_dB)

#%%
def _habitat_att_coeff_dB (f, a0=0.02):
  """ 
  get the habitat attenuation coefficient in dB/m for the frequency f knowing 
  the habitat attenuation parameter a0 (in dB/kHz/m)
  
  Parameters
  ----------
  f: scalar or array-like
      frequency in Hz
  a0 : scalar, optional, default is 0.02
      attenuation coefficient of the habitat in dB/kHz/m 
      
  Returns
  -------
  coeff.Ahab.dB  : scalar or array-like (vector (1D)) 
      habitat attenuation coefficient in dB/m for the frequency f
      
  """
  # make sure it's array
  f = np.asarray(f)  
  
  Ahab_coeff_dB = a0*f/1000 
  
  return Ahab_coeff_dB

#%%
def _habitat_att_coeff (f, a0=0.02):
  """ 
  get the habitat attenuation factor for the frequency f knowing the habitat 
  attenuation parameter a0 (in dB/kHz/m)
  
  Parameters
  ----------
  f: scalar or array-like
      frequency in Hz
  a0 : scalar, optional, default is 0.02
      attenuation coefficient of the habitat in dB/kHz/m 
      
  Returns
  -------
  coeff.Ahab : scalar or array-like (vector (1D)) 
      habitat attenuation factor for the frequency f
  """
  # make sure it's array
  f = np.asarray(f)  
  
  Ahab_coeff = _habitat_att_coeff_dB(f,a0) /20*log10(exp(1))
  return Ahab_coeff

#%%
def _attenuation_factor (f, r, r0, t=20, rh=60, pa=101325, a0=0.02) :
  """ 
  get attenuation factor taking into account the geometric, atmospheric 
  and habitat attenuation contributions
  
  Parameters
  ----------
  f: scalar or array-like
      frequency in Hz
  r : scalar or array-like
      propagation distances in m 
  r0 : scalar
      reference distance in m
  t: scalar, optional, default is 20
      temperature in °C
  rh: scalar, optional, default is 60
      relative humidity in % 
  pa: scalar, optional, default is 101325
      atmospheric pressure in Pa 
  a0 : scalar, optional, default is 0.02
      attenuation coefficient of the habitat in dB/kHz/m 
  
  Returns
  -------
  A_factor : scalar or array-like (vector (1D) or matrix (2D))
      attenuation depending on the frequency and the distance [MATRIX] 
      => Multiply this value with the effective reference acoustic pressure p0 
      measured at r0 to estimate the pressure after attenuation taking into 
      account 3 types of attenuation
      rows : frequencies
      columns :distances
  """
  # make sure it's array
  f = np.asarray(f)  
  r = np.asarray(r)
  
  Ageo_factor = _geometric_att_factor(r,r0)
  Aatm_factor = _atmospheric_att_factor(f,r,r0,t,rh,pa)
  Ahab_factor = _habitat_att_factor(f,r,r0,a0)
  
  # make sure it's array
  Ageo_factor = np.asarray(Ageo_factor)  
  Aatm_factor = np.asarray(Aatm_factor)
  Ahab_factor = np.asarray(Ahab_factor)

  A_factor = Ageo_factor[np.newaxis, ...] * Aatm_factor * Ahab_factor
  
  return A_factor

#%%
# =============================================================================
# Public functions
# =============================================================================
def attenuation_dB (f, r, r0, t=20, rh=60, pa=101325, a0=0.02):
  """ 
  Compute the attenuation in decibels taking into account the geometric, atmospheric 
  and habitat attenuation contributions.
  
  Parameters
  ----------
  f: scalar or array-like
      frequency in Hz
  r : scalar or array-like
      propagation distances in m 
  r0 : scalar
      reference distance in m
  t: scalar, optional, default is 20
      temperature in °C
  rh: scalar, optional, default is 60
      relative humidity in % 
  pa: scalar, optional, default is 101325
      atmospheric pressure in Pa 
  a0 : scalar, optional, default is 0.02
      attenuation coefficient of the habitat in dB/kHz/m 
  
  Returns
  -------
  Asum_dB : scalar or array-like (vector (1D) or matrix (2D))
      Global attenuation in decibel taking into account the geometric, atmospheric 
      and habitat attenuation.
      => subtract Asum_dB from the reference acoustic pressure L0 in dB 
      (or sound pressure level (SPL)) measuread at distance r0 for each 
      frequency and distance to estimate the pressure after attenuation taking 
      into account 3 types of attenuation
      rows : frequencies
      columns :distances
      
  Examples
  --------
  >>> import numpy as np
  >>> fn = np.arange(500,10500,500)
  >>> r = np.arange(10,110)
  >>> r0 = 1
  >>> Asum_dB, df_att = maad.spl.attenuation_dB(fn,r,r0)
  
  For 2kHz @50m, the attenuation is mainly drive by geometric attenuation
  
  >>> df_att[(df_att.f==2000) & (df_att.r==50)]
              f   r  Ageo_dB   Aatm_dB   Ahab_dB    Asum_dB
      310  2000  50  33.9794  0.454776      1.96  36.394176
  
  For 10Hz @100m, the attenuation due to the atmosphere and the habitat is no
  negligible
      
  >>> df_att[(df_att.f==10000) & (df_att.r==100)]
                f   r  Ageo_dB      Aatm_dB   Ahab_dB    Asum_dB
      1990  10000  100    40.0    13.335002      19.8  73.135002
  """
  # make sure it's array
  f = np.asarray(f)  
  r = np.asarray(r)
  
  Ageo_dB = _geometric_att_dB(r,r0)
  Aatm_dB = _atmospheric_att_dB(f,r,r0,t,rh,pa)
  Ahab_dB = _habitat_att_dB(f,r,r0,a0)
  
  # make sure it's array
  Ageo_dB = np.asarray(Ageo_dB)  
  Aatm_dB = np.asarray(Aatm_dB)
  Ahab_dB = np.asarray(Ahab_dB)
    
  Asum_dB = Ageo_dB[np.newaxis,...] + Aatm_dB  + Ahab_dB
  
  #### create a dataframe with all values
  # test if f is a single value
  if f.ndim == 0 : Nf= 1
  else: Nf = len(f)
  # test if r is a single value
  if r.ndim == 0 : Nr= 1
  else: Nr = len(r)
  
  # for Ageo, repeat vector along frequency axis
  Ageo_dB = np.tile(Ageo_dB,(Nf,1))
  # for f, repeat vector along distance axis
  f = np.tile(f,(Nr,1)).T
  # for r, repeat vector along frequency axis  
  r = np.tile(r,(Nf,1))
  
  # prepare data (transform matrices into long vectors)
  data = {'f':f.flatten(), 
          'r':r.flatten(), 
          'Ageo_dB':Ageo_dB.flatten(), 
          'Aatm_dB':Aatm_dB.flatten(), 
          'Ahab_dB':Ahab_dB.flatten(), 
          'Asum_dB':Asum_dB.flatten()}
  df_att = pd.DataFrame(data, columns = ['f', 
                                         'r', 
                                         'Ageo_dB', 
                                         'Aatm_dB', 
                                         'Ahab_dB',
                                         'Asum_dB'])
  
  return Asum_dB, df_att

#%%
def dBSPL_per_bin (L, f) :
  """
  Function to spread the sound pressure level (Energy in dB) along a frequency 
  vector (bins).
  
  Parameters
  ----------
  L : scalar
      Sound Pressure Level in dB
  f: array-like (vector (1D))
      frequency vector in Hz
  
  Returns
  -------
  L_per_bin :  array-like (vector (1D))
      sound pressure level in dB with values corresponding to the frequency's 
      number of bins 
  
  Examples
  --------
  Spread 80dB SPL from 2kHz to 5kHz with 1kHz step
  
  >>> maad.spl.dBSPL_per_bin(80,[2000,3000,4000,5000])
  array([73.97940009, 73.97940009, 73.97940009, 73.97940009])
  
  Spread 80dB SPL from 2kHz to 10kHz with 1kHz step
  
  >>> maad.spl.dBSPL_per_bin(80,[2000,3000,4000,5000,6000,7000,8000,9000,10000])
  array([70.45757491, 70.45757491, 70.45757491, 70.45757491, 70.45757491,
       70.45757491, 70.45757491, 70.45757491, 70.45757491])
  
  Spread 80dB SPL from 2kHz to 5kHz with 0.5kHz step
  
  >>> maad.spl.dBSPL_per_bin(80,[2000,2500,3000,3500,4000,4500,5000])
  array([71.5490196, 71.5490196, 71.5490196, 71.5490196, 71.5490196,
       71.5490196, 71.5490196])  
  
  """
  # test if f is a scalar
  if not hasattr(f, "__len__") : 
      L_per_bin = L 
  # if f is a vector
  else:
      # force to be ndarray
      f = np.asarray(f)   
      # init
      L_per_bin = np.ones(len(f)) * L
      nb_bin= len(f)
      # dB SPL for the frequency bandwidth
      L_per_bin = L_per_bin - 10*log10(nb_bin) 

  return L_per_bin

#%%
def detection_distance (L_bkg, L0, f, r0= 1, delta_r=1, t=20, rh=60, pa=101325, 
                     a0=0.02, rmax=10000):
  """ 
  Compute the detection distance also known as detection range or detection radius
  or active space.
  
  Parameters
  ----------
  L_bkg : scalar or array-like
      sound pressure level of the background/ambient "noise" in dB SPL
  L0 : scalar or array-like
      Initial sound pressure level measured at distance r0
  f: scalar or array-like
      frequency in Hz
  r0 : scalar
      distance at which L0 was measured (generally @1m)
  delta_r : scalar
      distance resolution in m 
  t: scalar, optional, default is 20
      temperature in °C
  rh: scalar, optional, default is 60
      relative humidity in % 
  pa: scalar, optional, default is 101325
      atmospheric pressure in Pa 
  a0 : scalar, optional, default is 0.02
      attenuation coefficient of the habitat in dB/kHz/m 
  rmax : scalar, optional, default is 10000
      define the maximal distance that is taken into account for active distance
      calculation. Default value is 10000m (i.e. 10km) which is OK for most of
      the purpose. If you increase the value, the calculation will be longer      
        
  Returns
  -------
  distance_max : scalar or array-like
      maximum distance of propagation before the sound pressure level is below 
      the ambient sound level
  
  Notes
  -----
  The maximum detection range is limited by the background noise or ambient sound 
  also called noise. The signal is attenuated during the propagation until
  the level reaches the ambient sound level, meaning that the signal could not 
  be disantangle from the background sound anymore.
  
  Examples
  --------
  Estimation of the active distance for an initial sound of 90dB SPL @2kHz 
  with a background noise of 50dB SPL
    
  >>> f, r = maad.spl.active_distance (L_bkg=50, L0=90, f=2000) 
  >>> print('Max active distance is %2.1fm' %r)
  Max active distance is 70.0m
  
  Estimation of the active distance for an initial sound of 90dB SPL @10kHz 
  with a background noise of 50dB SPL
    
  >>> f, r = maad.spl.active_distance (L_bkg=50, L0=90, f=10000) 
  >>> print('Max active distance is %2.1fm' %r)
  Max active distance is 32.0m
  
  Estimation of the active distance for an initial sound of 90dB SPL @2kHz 
  with a background noise of 30dB SPL
    
  >>> f, r = maad.spl.active_distance (L_bkg=30, L0=90, f=2000) 
  >>> print('Max active distance is %2.1fm' %r)
  Max active distance is 263.0m
  
  Estimation of the active distance for an initial sound of 90dB SPL @2kHz 
  with a background noise of 30dB SPL, in tropical rain forest atmosphere 
  (t=30, rh=99)
    
  >>> f, r = maad.spl.active_distance(L_bkg=30,L0=90,f=2000,t=30,rh=99) 
  >>> print('Max active distance is %2.1fm' %r)
  Max active distance is 247.0m
  
  Estimation of the active distance for an initial sound of 90dB SPL @2kHz 
  with a background noise of 30dB SPL, in cold dry forest atmosphere 
  (t=5, rh=40)
    
  >>> f, r = maad.spl.active_distance(L_bkg=30,L0=90,f=2000,t=5,rh=40) 
  >>> print('Max active distance is %2.1fm' %r)
  Max active distance is 225.0m
  
  """
    
  # test if f is a scalar
  if not hasattr(f, "__len__") : f = np.array([f])
  if not hasattr(L_bkg, "__len__") : L_bkg = np.array([L_bkg])
  if not hasattr(L0, "__len__") : L0 = np.array([L0])

  # test if f, L_bkg and L0 have the same length
  if not (len(f)==len(L_bkg) & len(f)==len(L_bkg) & len(L0)==len(L_bkg)) : 
      raise TypeError ('L_bkg, L0 and f must have the same length')  
      
  # test if r0 and a0 are scalars
  if hasattr(r0, "__len__") : raise TypeError ('r0 must be a scalar ') 
  if hasattr(rmax, "__len__") : raise TypeError ('rmax must be a scalar ')   

  # set the distance vector
  r = np.arange(1,rmax,delta_r) 

  # number of frequencies
  Nf = len(f)     
  
  # set the distance max vector to store the result   
  distance_max = np.zeros(Nf)

  # test for each frequency if the signal sound level at distance r is below
  # the background sound level    
  for ii in np.arange(Nf) :  
      # estimate the full attenuation of the signal
      Asum_dB, _ = attenuation_dB(f[ii], r, r0, t, rh, pa, a0)
      
      # Get the sound level after subtracting the attenuation
      L = L0[ii] - Asum_dB  
      
      # distance max that corresponds to a 
      if sum((L - L_bkg[ii])>0) >1 :
          distance_max[ii] = r[np.argmin((L - L_bkg[ii])[(L - L_bkg[ii])>0])] 
      else :
          distance_max[ii] = 0
      
  # test if f and distance_max are scalars
  if len(f) == 1 : f = f[0]
  if len(distance_max) == 1 : distance_max = distance_max[0]
  
  # return the frequency vector associated with the distance max
  return f,distance_max

#%%
def pressure_at_r0 (f, r, p, r0=1, t=20, rh=60, pa=101325, a0=0.02) :
  """ 
    Estimate the pressure p0 at distance r0 from pressure p measured at 
    distance r. This function takes into account the geometric, atmospheric 
    and habitat attenuations.

  Parameters
  ----------
  f: scalar or array-like
      frequency in Hz
  r: scalar or array-like
      distance in m at which p in Pa was measured 
  p : scalar or array-like
      pressure in Pa
  r0 : scalar
      distance at which p0 needs to be estimated (generally @1m)
  t: scalar, optional, default is 20
      temperature in °C
  rh: scalar, optional, default is 60
      relative humidity in % 
  pa: scalar, optional, default is 101325
      atmospheric pressure in Pa 
  a0 : scalar, optional, default is 0.02
      attenuation coefficient of the habitat in dB/kHz/m 
        
  Returns
  -------
  p0 : scalar or array-like
      estimated pressure at distance r0.
      This is a scalar if p is a scalar or a vector if p and r are vectors or 
      a matrix if p and r are matrix 

  Examples
  --------
  Estimation of L0, knowing L=50dB SPL @60m @100Hz
  
  >>> p = maad.spl.dBSPL2pressure(50)
  >>> p0 = maad.spl.pressure_at_r0(100,60,p)
  >>> L0 = maad.spl.pressure2dBSPL(p0)
  >>> L0
  85.68036510289447
  
  Estimation of L0, knowing L=50dB SPL @60m @10000Hz
  
  >>> p0 = maad.spl.pressure_at_r0(10000,100,p)
  >>> L0 = maad.spl.pressure2dBSPL(p0)
  >>> L0
  120.53306313162624
  
  Estimation of L0, knowing L=50dB SPL @10m @100Hz
  
  >>> p = maad.spl.dBSPL2pressure(50)
  >>> p0 = maad.spl.pressure_at_r0(100,10,p)
  >>> L0 = maad.spl.pressure2dBSPL(p0)
  >>> L0
  70.01789933655922
      
  """
  # if vectors, test if r and L have same length 
  if (hasattr(r, "__len__") and hasattr(p, "__len__")) :
      if not (len(r) == len(p)) :  raise TypeError ('if vectors, r and p must have the same length') 
  
  # if scalar set f, r L as ndarray 
  if not hasattr(f, "__len__") : f = np.array([f])
  if not hasattr(r, "__len__") : r = np.array([r])
  if not hasattr(p, "__len__") : p = np.array([p])
  
  # force to be an ndarray
  f = np.asarray(f)
  r = np.asarray(r)
  p = np.asarray(p)
    
  # f, r and L must be scalars or vectors
  if f.ndim>1 : raise TypeError ('f must be a scalar or a vector') 
  if r.ndim>1 : raise TypeError ('r must be a scalar or a vector')  
  if p.ndim>1 : raise TypeError ('p must be a scalar or a vector')  
  
  Ageo_factor = _geometric_att_factor(r,r0)
  Aatm_factor = _atmospheric_att_factor(f,r,r0,t,rh,pa)
  Ahab_factor = _habitat_att_factor(f,r,r0,a0)
  
  if Ageo_factor.size == Aatm_factor.size:
      p0 = p * Ageo_factor**(-1) * Aatm_factor**(-1) * Ahab_factor**(-1)
  else:    
      p0 = p * (Ageo_factor**(-1))[np.newaxis,...] * Aatm_factor**(-1) * Ahab_factor**(-1)
  
  if hasattr(p0, "__len__") :
      # reshape the array when dimensions are 1
      if p0.ndim ==1:
          if p0.shape[0] == 1:
              p0 = p0[0]  
  return p0  

#%%
def dBSPL_at_r0 (f, r, L, r0=1, t=20, rh=60, pa=101325, a0=0.02, pRef=10e-6) :
  """ 
    Estimate the sound pressure level L0 (dB SPL) at distance r0 from sound pressure
    level L measured at distance r.
    
    This function takes into account the geometric, atmospheric and habitat attenuations.

  Parameters
  ----------
  f: scalar or array-like
      frequency in Hz
  r: scalar or array-like
      distance in m at which p in Pa was measured 
      if r is a vector, r must have the same length as L
  L : scalar or array-like
      Sound pressure level L in dB SPL
      if L is a vector, L must have the same length as r
  r0 : scalar
      distance at which p0 needs to be estimated (generally @1m)
  t: scalar, optional, default is 20
      temperature in °C
  rh: scalar, optional, default is 60
      relative humidity in % 
  pa: scalar, optional, default is 101325
      atmospheric pressure in Pa 
  a0 : scalar, optional, default is 0.02
      attenuation coefficient of the habitat in dB/kHz/m 
        
  Returns
  -------
  L0 : scalar or array-like
      estimated sound presssure level L0 at distance r0 [SCALAR]
  """ 
  
  # Transform the pressure into dB SPL
  p = dBSPL2pressure(L)  
  
  # Get the initial pressure (Pa)
  p0 = pressure_at_r0(f, r, p, r0, t, rh, pa, a0)
  
  # Transform the pressure into dB SPL
  L0 = pressure2dBSPL(p0, pRef)
  
  return L0  

#%%
def apply_attenuation (p0, fs, r, r0= 1, t=20, rh=60, pa=101325, a0=0.02):
  """ 
  Apply attenuation of a temporal signal p0 after propagation between the 
  reference distance r0 and the final distance r taken into account the 
  geometric, atmospheric and habitat attenuation contributions.
  
  Parameters
  ----------
  p0 : array-like (vector 1d)
      temporal signal p0 in Pa (time domain)
  fs: scalar 
      sampling frequency Hz
  r: scalar
      distance of propagation (in m) of the signal p0 
  r0 : scalar
      distance at which the temporal signal p0 was measured
  t: scalar, optional, default is 20
      temperature in °C
  rh: scalar, optional, default is 60
      relative humidity in % 
  pa: scalar, optional, default is 101325
      atmospheric pressure in Pa 
  a0 : scalar, optional, default is 0.02
      attenuation coefficient of the habitat in dB/kHz/m 
  
  Returns
  -------
  p :  array-like (vector 1d)
      temporal signal (time domain) after attenuation
  
  Examples
  --------
  Prepare the spinetail sound (Sound level @1m = 85dB SPL).
  
  >>> w, fs = maad.sound.load('../data/spinetail.wav') 
  >>> p0 = maad.spl.wav2pressure(wave=w, gain=42)
  >>> p0_sig = p0[int(5.68*fs):int(7.48*fs)] 
  >>> p0_noise = p0[int(8.32*fs):int(10.12*fs)] 
  >>> Sxx_power, tn, fn, ext = maad.sound.spectrogram(p0_sig ,fs)
  >>> Sxx_power_noise, tn, fn, ext = maad.sound.spectrogram(p0_noise ,fs)
  >>> Sxx_dB = maad.util.power2dB(Sxx_power, db_range=96) + 96
  >>> Sxx_dB_noise = maad.util.power2dB(Sxx_power_noise, db_range=96) + 96
  
  Get the sound level of the spinetail song (sound between 4900-7500 Hz).
  
  >>> p0_sig_4900_7500 = maad.sound.select_bandwidth(p0_sig,fs,fcut=[4900,7300],forder=10, ftype='bandpass')
  >>> L = maad.spl.pressure2leq(p0_sig_4900_7500, fs) 
  >>> print ('Sound Level measured : %2.2fdB SPL' %L)
  
  Estimate maximum distance from the source.
  
  >>> r = maad.spl.active_distance(L, 85, f=(7500+4900)/2) 
  
  plot original spectrogram
  
  >>> import matplotlib.pyplot as plt
  >>> fig, (ax1, ax2, ax3, ax4, ax5) = plt.subplots(1,5, sharex=True, figsize=(15,3))
  >>> maad.util.plot2d(Sxx_dB, ax=ax1, extent=ext, vmin=0, vmax=70, figsize=[3,3])

  Compute the audio attenuation at 10m.
  
  >>> p_att = maad.spl.apply_attenuation(p0_sig, fs, r0=5, r =10)
  >>> Sxx_power_att, tn, fn, ext = maad.sound.spectrogram(p_att,fs)
  >>> Sxx_dB_att_10m = maad.util.power2dB(Sxx_power_att,db_range=96) + 96 
  
  >>> p_att = maad.spl.apply_attenuation(p0_sig, fs, r0=5, r =20)
  >>> Sxx_power_att, tn, fn, ext = maad.sound.spectrogram(p_att,fs)
  >>> Sxx_dB_att_20m = maad.util.power2dB(Sxx_power_att,db_range=96) + 96 
  
  >>> p_att = maad.spl.apply_attenuation(p0_sig, fs, r0=5, r =40)
  >>> Sxx_power_att, tn, fn, ext = maad.sound.spectrogram(p_att,fs)
  >>> Sxx_dB_att_40m = maad.util.power2dB(Sxx_power_att,db_range=96) + 96 
  
  >>> p_att = maad.spl.apply_attenuation(p0_sig, fs, r0=5, r =80)
  >>> Sxx_power_att, tn, fn, ext = maad.sound.spectrogram(p_att,fs)
  >>> Sxx_dB_att_80m = maad.util.power2dB(Sxx_power_att,db_range=96) + 96 
  
  Add noise to the signal.
  
  >>> Sxx_dB_att_10m = maad.util.add_dB(Sxx_dB_att_10m,Sxx_dB_noise) - 3 
  >>> Sxx_dB_att_20m = maad.util.add_dB(Sxx_dB_att_20m,Sxx_dB_noise) - 3 
  >>> Sxx_dB_att_40m = maad.util.add_dB(Sxx_dB_att_40m,Sxx_dB_noise) - 3 
  >>> Sxx_dB_att_80m = maad.util.add_dB(Sxx_dB_att_80m,Sxx_dB_noise) - 3 
    
  Plot attenuated spectrogram.
  
  >>> maad.util.plot2d(Sxx_dB_att_10m, ax=ax2, extent=ext, vmin=0, vmax=70, figsize=[3,3])
  >>> maad.util.plot2d(Sxx_dB_att_20m, ax=ax3, extent=ext, vmin=0, vmax=70, figsize=[3,3])
  >>> maad.util.plot2d(Sxx_dB_att_40m, ax=ax4, extent=ext, vmin=0, vmax=70, figsize=[3,3])
  >>> maad.util.plot2d(Sxx_dB_att_80m, ax=ax5, extent=ext, vmin=0, vmax=70, figsize=[3,3])     
  
  """
  
  # Fourier domain
  p0_f = np.fft.fft(p0)
  f = np.arange(len(p0_f)) / len(p0_f) * fs /2
  # apply attenuation
  p_f = p0_f * _attenuation_factor(f, r, r0, t, rh, pa, a0)
  # Go back to the time domain
  p = np.fft.ifft(p_f)
  # keep the real part
  p = np.real(p)
  
  return (p)



